import numpy as np
from scipy.optimize import minimize

import deampy.random_variates as RVG
import deampy.statistics as Stat


def assert_np_list(obs, error_message):
    """
    :param obs: list of observations to convert to np.array
    :param error_message: error message to display if conversion failed
    :return: np.array of obs
    """

    if type(obs) is not list and type(obs) is not np.ndarray:
        obs = [obs]

    try:
        new_array = np.array(obs)
    except ValueError:
        raise ValueError(error_message)

    return new_array


def inmb_u(d_effect, d_cost):
    """ higher d_effect represents better health """
    return lambda w: w*d_effect - d_cost


def inmb_d(d_effect, d_cost):
    """ higher d_effect represents worse health """
    return lambda w: -w * d_effect - d_cost


def inmb2_u(d_effect, d_cost):
    return lambda w_gain, w_loss: \
        inmb_u(d_effect, d_cost)(w_gain) if d_effect >= 0 else inmb_u(d_effect, d_cost)(w_loss)


def inmb2_d(d_effect, d_cost):
    return lambda w_gain, w_loss: \
        inmb_d(d_effect, d_cost)(w_gain) if d_effect >= 0 else inmb_d(d_effect, d_cost)(w_loss)


def get_d_cost(strategy):
    # this is defined for sorting strategies
    return strategy.dCost.get_mean()


def get_d_effect(strategy):
    # this is defined for sorting strategies
    return strategy.dEffect.get_mean()


def get_index(strategy):
    # this is defined for sorting strategies
    return strategy.idx


def find_intersecting_wtp(w0, u_new, u_base):
    # to find the intersecting WTP threshold for a general utility function

    if u_new(w0) > u_base(w0):
        return None

    else:
        f = lambda w: (u_new(w)-u_base(w))**2
        res = minimize(f, (w0))
        w_star = res.x[0]

        if abs(u_new(w_star)-u_base(w_star)) > 0.01:
            return None

        if w_star > w0:
            return w_star
        else:
            return None


def utility_sample_stat(utility, d_cost_samples, d_effect_samples,
                        wtp_random_variate, n_samples, rnd):

    discrete_rnd = RVG.UniformDiscrete(
        l=0, u=len(d_cost_samples)-1)

    samples = []
    for i in range(n_samples):
        j = discrete_rnd.sample(rnd)

        u = utility(d_effect=d_effect_samples[j],
                    d_cost=d_cost_samples[j])

        w = wtp_random_variate.sample(rnd)
        samples.append(u(w))

    return Stat.SummaryStat(name='', data=samples)


def update_curves_with_highest_values(wtp_values, curves):
    """ find strategies with the highest expected incremental net monetary benefit.
        and return the indices of these strategies over the range of wtp values
     """

    # find the optimal strategy for each wtp value
    idx_highest_exp_value = []
    for wtp_idx, wtp in enumerate(wtp_values):

        max_value = float('-inf')
        max_idx = 0
        for s_idx in range(len(curves)):
            if curves[s_idx].ys[wtp_idx] > max_value:
                max_value = curves[s_idx].ys[wtp_idx]
                max_idx = s_idx

        # store the index of the optimal strategy
        idx_highest_exp_value.append(max_idx)
        curves[max_idx].frontierXs.append(wtp)
        curves[max_idx].frontierYs.append(max_value)

    for curve in curves:
        curve.convert_lists_to_arrays()

    return idx_highest_exp_value


def update_curves_with_lowest_values(wtp_values, curves):
    """ find strategies with the lowest expected loss.
        and return the indices of these strategies over the range of wtp values
     """

    # find the optimal strategy for each wtp value
    idx_lowest_exp_value = []
    for wtp_idx, wtp in enumerate(wtp_values):

        min_value = float('inf')
        min_idx = 0
        for s_idx in range(len(curves)):
            if curves[s_idx].ys[wtp_idx] < min_value:
                min_value = curves[s_idx].ys[wtp_idx]
                min_idx = s_idx

        # store the index of the optimal strategy
        idx_lowest_exp_value.append(min_idx)
        curves[min_idx].frontierXs.append(wtp)
        curves[min_idx].frontierYs.append(min_value)

    for curve in curves:
        curve.convert_lists_to_arrays()

    return idx_lowest_exp_value


class _Curve:
    def __init__(self, label, color, linestyle, short_label):
        self.label = label
        self.shortLabel = label if short_label is None else short_label
        self.color = color
        self.linestyle = linestyle
        self.xs = []
        self.ys = []
        self.frontierXs = []
        self.frontierYs = []
        self.l_errs = None  # lower error length of health outcome over a range of budget values
        self.u_errs = None  # upper error length of health outcome over a range of budget values

    def convert_lists_to_arrays(self):
        self.xs = np.array(self.xs)
        self.ys = np.array(self.ys)
        self.frontierXs = np.array(self.frontierXs)
        self.frontierYs = np.array(self.frontierYs)


class INMBCurve(_Curve):
    # incremental net monetary benefit curve of one strategy
    def __init__(self, label, color, wtp_values, inmb_stat, interval_type='n', short_label=None):
        """
        :param label: (string) label of this incremental NMB curve
        :param color: color code of this curve
        :param wtp_values: wtp values over which this curve should be calcualted
        :param inmb_stat: incremental NMB statistics
        :param interval_type: (string) 'n' for no interval
                                       'c' for t-based confidence interval,
                                       'p' for percentile interval
        :param short_label: (string) to display on the curves
        """

        _Curve.__init__(self=self, label=label, color=color, short_label=short_label, linestyle='-')
        self.inmbStat = inmb_stat
        self.xs = wtp_values
        self.intervalType = interval_type

        self._calculate_ys_lerrs_uerrs()

    def _calculate_ys_lerrs_uerrs(self):
        """
        calculates the expected incremental NMB and the confidence or prediction intervals over the specified
        range of wtp values.
        """

        self.l_errs = []
        self.u_errs = []

        # get the NMB values for each wtp
        self.ys = [self.inmbStat.get_incremental_nmb(x) for x in self.xs]

        if self.intervalType == 'c':
            y_intervals = [self.inmbStat.get_CI(x, alpha=0.05) for x in self.xs]
        elif self.intervalType == 'p':
            y_intervals = [self.inmbStat.get_PI(x, alpha=0.05) for x in self.xs]
        elif self.intervalType == 'n':
            y_intervals = None
        else:
            raise ValueError('Invalid value for internal_type.')

        # reshape confidence interval to plot
        if y_intervals is not None:
            self.u_errs = np.array([p[1] for p in y_intervals]) - self.ys
            self.l_errs = self.ys - np.array([p[0] for p in y_intervals])
        else:
            self.u_errs, self.l_errs = None, None
    
    def get_switch_wtp(self):
        
        return self.inmbStat.get_switch_wtp()
    
    def get_switch_wtp_and_interval(self):

        return self.inmbStat.get_switch_wtp_and_interval(
            wtp_range=[self.xs[0], self.xs[-1]],
            interval_type=self.intervalType)    


class AcceptabilityCurve(_Curve):
    """ cost-effectiveness acceptability curve of one strategy """

    def __init__(self, label, color, short_label=None):

        _Curve.__init__(self, label=label, color=color, short_label=short_label, linestyle='-')


class ExpectedLossCurve(_Curve):
    """ expected loss curve of one strategy """

    def __init__(self, label, color, short_label=None):

        _Curve.__init__(self, label=label, color=color, short_label=short_label, linestyle='-')


class ExpHealthCurve(_Curve):

    def __init__(self, label, color, effect_stat, interval_type='n', short_label=None):
        """
        :param label: (string) label of this incremental NMB curve
        :param color: color code of this curve
        :param effect_stat: statistics of health outcomes
        :param interval_type: (string) 'n' for no interval
                                       'c' for t-based confidence interval,
                                       'p' for percentile interval
        :param short_label: (string) to display on the curves
        """

        _Curve.__init__(self, label=label, color=color, short_label=short_label, linestyle='-')

        self.dEffectMean = effect_stat.get_mean()

        if interval_type == 'c':
            interval = effect_stat.get_t_CI(alpha=0.05)
        elif interval_type == 'p':
            interval = effect_stat.get_PI(alpha=0.05)
        elif interval_type == 'n':
            interval = None
        else:
            raise ValueError('Invalid value for internal_type.')

        self.l_errs = []
        self.u_errs = []
        if interval:
            self.l_err = effect_stat.get_mean() - interval[0]
            self.u_err = interval[1] - effect_stat.get_mean()
        else:
            self.l_err = None
            self.u_err = None

    def update_feasibility(self, b):
        self.xs.append(b)
        self.ys.append(self.dEffectMean)
        self.l_errs.append(self.l_err)
        self.u_errs.append(self.u_err)


class EVPI(_Curve):
    """ curve of expected value of perfect information """

    def __init__(self, xs, ys, label, color, short_label=None):

        _Curve.__init__(self, label=label, color=color, short_label=short_label, linestyle='dashdot')

        self.xs = assert_np_list(obs=xs, error_message='x-values should be list or numpy.array')
        self.ys = assert_np_list(obs=ys, error_message='y-values should be list of numpy.array')






def index_of_an_action_combo(action_combo):

    s = 0
    n = len(action_combo)
    for i in range(n):
        s += action_combo[i] * pow(2, n-i-1)
    return int(s)


def action_combo_of_an_index(index):

    string = bin(index).replace("0b", "")
    return list(map(int, string))



import deampy.random_variates as RVGs
from deampy.data_structure import DataFrameOfObjects, OneDimDataFrame


class MortalityModelByAgeSex:
    """
    example:
        age,   sex,      mortality rate
        0,     0,        0.1,
        0,     1,        0.11,
        5,     0,        0.2,
        5,     1,        0.21,
        10,    0,        0.3
        10,    1,        0.31

    This class assumes that the first column contains age groups and the last column contains mortality rates
    """

    def __init__(self, mortality_rates, group_mins, group_maxs, group_delta, age_min, age_delta):
        """
        :param mortality_rates: (list of list) the table above
        :param group_mins: list of minimum value of x (in example above: [0, 0])
        :param group_maxs: list of maximum value of x (in example above: [10, 1])
        :param group_delta: list of interval between break points of x
                    if set to 'int', x is treated as categorical variable
                    (in example above: [5, 'int'])
        :param age_min:
        :param age_delta:
        """

        self.df = DataFrameOfObjects(list_x_min=group_mins,
                                     list_x_max=group_maxs,
                                     list_x_delta=group_delta)

        self.ageMin = age_min

        for df_row in self.df.get_rows():
            rates = []
            for row in mortality_rates:
                if df_row[:-1] == row[1:-1]:
                    rates.append(row[-1])
            self.df.set_obj(x_value=df_row[0:-1],
                            obj=RVGs.NonHomogeneousExponential(rates=rates, delta_t=age_delta))

    def sample_time_to_death(self, group, age, rng):

        if age < self.ageMin:
            raise ValueError('Current age cannot be smaller than the minimum age.')

        return self.df.get_obj(group).sample(rng=rng, arg=age)



class MortalityModelByAge:
    """
    example:
        age,   mortality rate
        0,     0.1,
        5,     0.2,
        10,    0.3
    """

    def __init__(self, age_breaks, mortality_rates, age_delta):
        """
        :param : (list of list) the table above
        :param age_delta:
        """

        if len(age_breaks) != len(mortality_rates):
            raise ValueError('The number of age breaks should be equal to the number of mortality rates.')

        # create the list of nonhomogenous exponential distributions
        y_objects = []
        for i, v in enumerate(mortality_rates):
            if v <= 0:
                raise ValueError('All y_values (rates of exponential distributions) should be greater than 0.')
            y_objects.append(RVGs.NonHomogeneousExponential(
                rates= mortality_rates[i:]
            ))

        self.ageMin = age_breaks[0]

        self.df = OneDimDataFrame(
            y_objects=y_objects,
            x_min=self.ageMin,
            x_max=age_breaks[-1],
            x_delta=age_delta)

    def sample_time_to_death(self, current_age, rng):

        if current_age < self.ageMin:
            raise ValueError('Current age cannot be smaller than the minimum age.')

        return self.df.get_obj(current_age).sample(rng=rng)



import matplotlib.pyplot as plt
import numpy as np


class SimModel:
    # abstract class to be overridden by the simulation model to optimize
    def __init__(self):
        pass

    def get_obj_value(self, x, seed_index=0):
        """
        abstract method to return one realization of the objective function to optimize
        :param x: the values of the variables
        :param seed_index: specify if need to use a different seed for this simulation replication
        """

        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")


class StepSize_GeneralizedHarmonic:
    # generalized harmonic(GH) stepsize
    # step_n = a0 * b / (b + i), for i >= 0, a0 > 0, and b >= 1
    # (i is the iteration of the optimization algorithm)

    def __init__(self, a0, b=1):
        self._a0 = a0
        self._b = b

    def get_value(self, itr):
        return self._a0 * self._b / (itr + self._b)


class StepSize_Df:
    # step size: c/i^(-0.25) where c>0 and i>=0 is the iteration of the optimization_figs algorithm
    def __init__(self, c0):
        self._c0 = c0

    def get_value(self, itr):
        return self._c0 * pow(itr + 1, -0.25)


class StochasticApproximation:
    # stochastic approximation algorithm

    def __init__(self, sim_model,
                 derivative_step=StepSize_Df(c0=1),
                 step_size=StepSize_GeneralizedHarmonic(a0=1)
                 ):
        """
        :param sim_model: the simulation model to optimize
        :param derivative_step: derivative step if calculating slopes
        :param step_size: the step size rule
        """
        self._simModel = sim_model
        self._stepSize = step_size
        self._derivativeStep = derivative_step
        self.xStar = None   # optimal x value
        self.fStar = None   # optimal objective function 

        self.itr_i = []     # iteration indices
        self.itr_x = []     # x values over iterations
        self.itr_nDf = []    # normalized derivatives of f over iterations
        self.itr_f = []     # f values over iterations
        self.itr_stepmove=[] # moving steps over iterations
        self.itr_stepdf=[] # harmonic step size over iterations

    def minimize(self, max_itr, n_last_itrs_to_ave, x0):
        """
        :param max_itr: maximum iteration to terminate the algorithm
        :param n_last_itrs_to_ave: the number of last iterations to average as estimates of optimal x and f(x)
        :param x0: (list or numpy.array) starting point
        """

        # x0 has to be of type numpy.array
        if type(x0) is not np.ndarray:
            try:
                x0 = np.array(x0)
            except:
                raise ValueError('x0 should be a list or a numpy.array.')

        x = x0

        # iterations of stochastic approximation algorithm
        for itr in range(0, max_itr):

            # generate an array of numpy.array to calculate derivative
            # epsilon_matrix =
            #   [
            #       [ε, 0, 0, 0],
            #       [0, ε, 0, 0],
            #       [0, 0, ε, 0],
            #       [0, 0, 0, ε],
            #   ]

            step_df = self._derivativeStep.get_value(itr)
            step_move = self._stepSize.get_value(itr)

            epsilon_matrix = []
            for i in range(0, len(x)):
                # create an all zero array
                epsilon_array = np.zeros(len(x))
                # set ith element to epsilon (derivative step)
                epsilon_array[i] = step_df
                # append the array to the epsilon matrix
                epsilon_matrix.append(epsilon_array)

            # estimate the derivative of f at current x
            Df = np.array([])
            for i in range(0, len(x)):
                # partial derivative of variable i
                partial_derivative_i \
                    = (self._simModel.get_obj_value(x + epsilon_matrix[i], seed_index=itr) -
                       self._simModel.get_obj_value(x - epsilon_matrix[i], seed_index=itr))/(2*step_df)

                # append this partial derivative
                Df = np.append(Df, [partial_derivative_i])

            # normalize derivative
            norm = np.linalg.norm(Df, 2)
            if norm == 0:
                nDf = Df
            else:
                nDf = Df / np.linalg.norm(Df, 2)

            # evaluate the model at x
            f = self._simModel.get_obj_value(x, seed_index=itr)

            # store information at this iteration
            self.itr_i.append(itr)
            self.itr_x.append(x)
            self.itr_f.append(f)
            self.itr_nDf.append(nDf)
            self.itr_stepdf.append(step_df)
            self.itr_stepmove.append(step_move)

            # find a new x: x_new = x - step_size*f'(x)/||f'(x)||
            x = x - step_move * nDf

        # use the last n iterations to calculate optimal x and optimal objective value
        length = len(self.itr_x) - n_last_itrs_to_ave
        x_cal = self.itr_x[length:]
        f_cal = self.itr_f[length:]

        # store the optimal x and optimal objective value
        self.xStar = sum(x_cal)/len(x_cal)
        self.fStar = sum(f_cal)/len(x_cal) # should be the average of last f's

    def plot_f_itr(self, f_star=None):
        """
        :param f_star: optimal f value if known
        :return: a plot of f values as the algorithm iterates
        """

        fig, ax = plt.subplots(figsize=(6, 5))
        ax.plot(self.itr_i, self.itr_f)

        plt.xlabel('Iteration')
        plt.ylabel('Objective Function')
        if f_star is not None:
            plt.axhline(y=f_star, linestyle='--', color='black', linewidth=1)
        plt.show()

    def plot_x_irs(self, x_stars=None):
        """
        :param x_star: (list or numpy.array) optimal x value
        :return: plots of x_i in x as the the algorithm iterates
        """

        n_vars = len(self.itr_x[0])
        f, axarr = plt.subplots(n_vars, 1, sharex=True)

        for i in range(n_vars):
            # find x_i (ith dimension of x) over iterations
            x_i_itr = []
            for itr in range(len(self.itr_i)):
                x_i_itr.append(self.itr_x[itr][i])

            axarr[i].plot(self.itr_i, x_i_itr)  # , color=ser.color, alpha=0.5)
            axarr[i].set(ylabel='x' + str(i))

            if x_stars is not None:
                axarr[i].axhline(y=x_stars[i], linestyle='--', color='black', linewidth=1)

        # label the x-axis of the last figure
        axarr[n_vars-1].set(xlabel='Iteration')

        plt.show()

    def plot_Df_irs(self):
        """
        :return: plot of how the derivative is changing over iterations.
        """
        n_vars=len(self.itr_nDf[0])
        fig, ax = plt.subplots(n_vars, 1, sharex=True)

        for i in range(n_vars):
            # find x_i (ith dimension of x) over iterations
            df_i_itr = []
            for itr in range(len(self.itr_i)):
                df_i_itr.append(self.itr_nDf[itr][i])

            ax[i].plot(self.itr_i, df_i_itr)  # , color=ser.color, alpha=0.5)
            ax[i].set(ylabel='nDf' + str(i))
        # label the x-axis of nDf
        ax[n_vars - 1].set(xlabel='Iteration')

        plt.show()

    def plot_step_move(self):
        """
        :return: plot of how the harmonic step size are changing over iterations.
        """
        plt.plot(self.itr_i,self.itr_stepmove)
        plt.ylabel('harmonic step size')
        plt.xlabel('iterations')
        plt.show()

    def plot_step_Df(self):
        """
        :return: plot of how the derivative step size are changing over iterations.
        """
        plt.plot(self.itr_i,self.itr_stepdf)
        plt.ylabel('derivative step size')
        plt.xlabel('iterations')
        plt.show()


def plot_step_size(a0s, bs, c0s, nItrs):
      
    fig, ax = plt.subplots(figsize=(6, 5))
    ax.set_xlabel('Iteration')
    
    for a0 in a0s:
        for b in bs:
            GH = StepSize_GeneralizedHarmonic(a0=a0, b=b)
            
            y = []
            for itr in range(nItrs+1):
                y.append(GH.get_value(itr=itr))

            ax.plot(range(nItrs+1), y, label='GH: a0={}, b={}'.format(a0, b))

    for c0 in c0s:
        Df = StepSize_Df(c0=c0)
        y = []
        for itr in range(nItrs + 1):
            y.append(Df.get_value(itr=itr))

        ax.plot(range(nItrs + 1), y, linestyle='dashed', linewidth=3, label='Df: c0={}'.format(c0))
    
    fig.legend()
    fig.show()



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stat

import deampy.plots.plot_support as Fig
import deampy.random_variates as RVGs

COLOR_CONTINUOUS_FIT = 'r'
COLOR_DISCRETE_FIT = 'r'
MIN_PROP = 0.001
MAX_PROB = 0.999
MAX_PROB_DISCRETE = 0.999999
LEGEND_FONT_SIZE = 8
LINE_WIDTH = 2


def find_bins(data, bin_width):
    """
    :param data: (list) of observations
    :param bin_width: desired bin width
    :return: 'auto' if bin_width is not provided; otherwise, it calculates the binds
    """
    if bin_width is None:
        bins = 'auto'
    else:
        bins = np.arange(min(data), max(data) + bin_width, bin_width)
    return bins


def add_hist(ax, data, bin_width):
    """ add the histogram of the provided data to the axis"""
    ax.hist(data, density=True, bins=find_bins(data, bin_width),
            edgecolor='black', alpha=0.5, label='Data')


def add_continuous_dist(ax, dist, label, color=None):
    """ add the distribution of the provided continuous probability distribution to the axis
    :param ax: figure axis
    :param dist: probability distribution
    :param label: label of the fitted probability distribution to used in the legend
    :param color: color code of this distribution
    """

    x_values = np.linspace(dist.ppf(MIN_PROP),
                           dist.ppf(MAX_PROB), 200)
    ax.plot(x_values, dist.pdf(x_values),
            color=COLOR_CONTINUOUS_FIT if color is None else color,
            lw=LINE_WIDTH, label=label)


def add_discrete_dist(ax, dist, label, color=None):
    """ add the distribution of the provided discrete probability distribution to the axis
    :param ax: figure axis
    :param dist: probability distribution
    :param label: label of the fitted probability distribution to used in the legend
    :param color: color code of this distribution
    """

    x_values = range(int(dist.ppf(MIN_PROP)), int(dist.ppf(MAX_PROB))+1, 1)

    # probability mass function needs to be shifted to match the histogram
    pmf = dist.pmf(x_values)
    pmf = np.append(pmf[0], pmf[:-1])

    ax.step(x_values, pmf,
            color=COLOR_CONTINUOUS_FIT if color is None else color,
            lw=LINE_WIDTH, label=label)


def format_fig(ax, title, x_label, x_range, y_range):
    """ adds title, x_label, x_range, and y_range (it removes the ticks and lables of y-axis """

    ax.set_title(title)
    ax.set_xlim(x_range)
    ax.set_ylim(y_range)
    ax.yaxis.set_major_formatter(mpl.ticker.StrMethodFormatter('{x:,.0%}'))
    ax.set_xlabel(x_label)
    ax.set_yticks([])
    ax.set_yticklabels([])
    # ax.set_ylabel("Frequency")


def finish_ax(ax, data, bin_width, title, x_label, x_range, y_range, filename=None):
    """ add the histogram and format the figure """

    if data is not None:
        add_hist(ax, data, bin_width)

    # format axis
    format_fig(ax=ax, title=title, x_label=x_label, x_range=x_range, y_range=y_range)
    ax.legend(fontsize=LEGEND_FONT_SIZE)


def plot_continuous_dists(data, dists, labels, colors, title=None, x_label=None, x_range=None, y_range=None,
                          fig_size=(6, 5), bin_width=None, filename=None):

    # plot histogram
    fig, ax = plt.subplots(1, 1, figsize=fig_size)

    # plot the distributions
    for dist, label, color in zip(dists, labels, colors):
        add_continuous_dist(ax, dist, label=label, color=color)

    # add histogram and format
    finish_ax(ax=ax, data=data, bin_width=bin_width,
              title=title, x_label=x_label, x_range=x_range, y_range=y_range,
              filename=filename)

    Fig.output_figure(plt=fig, file_name=filename, dpi=300)


def plot_continuous_dists_in_row(data, dists, labels, colors, x_label=None, x_range=None, y_range=None,
                                 fig_size=(6, 5), bin_width=None, filename=None):

    # plot histogram
    fig, axes = plt.subplots(1, len(dists), figsize=fig_size)

    # plot the distributions
    for i in range(len(dists)):
        add_continuous_dist(axes[i], dists[i], label=labels[i], color=colors[i])

        # add histogram and format
        finish_ax(ax=axes[i], data=data, bin_width=bin_width,
                  title=None, x_label=x_label, x_range=x_range, y_range=y_range,
                  filename=filename)

    Fig.output_figure(plt=fig, file_name=filename, dpi=300)


def plot_discrete_dists(data, dists, labels, colors, title=None, x_label=None, x_range=None, y_range=None,
                        fig_size=(6, 5), bin_width=None, filename=None):

    # plot histogram
    fig, ax = plt.subplots(1, 1, figsize=fig_size)

    # plot the distributions
    for dist, label, color in zip(dists, labels, colors):
        add_discrete_dist(ax, dist, label=label, color=color)

    # add histogram and format
    finish_ax(ax=ax, data=data, bin_width=bin_width,
              title=title, x_label=x_label, x_range=x_range, y_range=y_range,
              filename=filename)

    Fig.output_figure(plt=fig, file_name=filename, dpi=300)


def plot_discrete_dists_in_row(data, dists, labels, colors, x_label=None, x_range=None, y_range=None,
                               fig_size=(6, 5), bin_width=None, filename=None):

    # plot histogram
    fig, axes = plt.subplots(1, len(dists), figsize=fig_size)

    # plot the distributions
    for i in range(len(dists)):
        add_discrete_dist(axes[i], dists[i], label=labels[i], color=colors[i])

        # add histogram and format
        finish_ax(ax=axes[i], data=data, bin_width=bin_width,
                  title=None, x_label=x_label, x_range=x_range, y_range=y_range,
                  filename=filename)

    Fig.output_figure(plt=fig, file_name=filename, dpi=300)


def plot_fit_continuous(data, dist, label, title=None, x_label=None, x_range=None, y_range=None,
                        fig_size=(6, 5), bin_width=None, filename=None):
    """ plots the pdf of a continuous probability distribution a long with the histogram of data. """

    fig, ax = plt.subplots(1, 1, figsize=fig_size)

    # plot the distribution
    add_continuous_dist(ax, dist, label=label)

    # add histogram and format
    finish_ax(ax=ax, data=data, bin_width=bin_width,
              title=title, x_label=x_label, x_range=x_range, y_range=y_range,
              filename=filename)

    Fig.output_figure(plt=fig, file_name=filename, dpi=300)


def plot_fit_discrete(data, dist, label, title=None, x_label=None, x_range=None, y_range=None,
                      fig_size=(6, 5), bin_width=None, filename=None):
    """ plots the pmf of a discrete probability distribution a long with the histogram of data. """

    fig, ax = plt.subplots(1, 1, figsize=fig_size)

    # plot the distribution
    add_discrete_dist(ax, dist, label=label)

    # add histogram and format
    finish_ax(ax=ax, data=data, bin_width=bin_width,
              title=title, x_label=x_label, x_range=x_range, y_range=y_range,
              filename=filename)

    Fig.output_figure(plt=fig, file_name=filename, dpi=300)


# ---- return distributions based on fit results ----
def get_beta_dist(fit_results):
    return stat.beta(fit_results['a'], fit_results['b'], fit_results['loc'], fit_results['scale'])


def get_beta_binomial_dist(fit_results):
    return stat.betabinom(n=fit_results['n'], a=fit_results['a'], b=fit_results['b'], loc=fit_results['loc'])


def get_binomial_dist(fit_results):
    return stat.binom(n=int(fit_results['n']), p=fit_results['p'], loc=fit_results['loc'])


def get_exponential_dist(fit_results):
    return stat.expon(fit_results['loc'], fit_results['scale'])


def get_gamma_dist(fit_results):
    return stat.gamma(fit_results['shape'], fit_results['loc'], fit_results['scale'])


def get_gamma_poisson_dist(fit_results):
    return RVGs.GammaPoisson(shape=fit_results['shape'], scale=fit_results['scale'], loc=fit_results['loc'])


def get_geometric_dist(fit_results):
    return stat.geom(p=fit_results['p'], loc=fit_results['loc'])


def get_johnson_sb_dist(fit_results):
    return stat.johnsonsb(a=fit_results['a'], b=fit_results['b'], loc=fit_results['loc'], scale=fit_results['scale'])


def get_johnson_su_dist(fit_results):
    return stat.johnsonsu(a=fit_results['a'], b=fit_results['b'], loc=fit_results['loc'], scale=fit_results['scale'])


def get_lognormal_dist(fit_results):
    return stat.lognorm(s=fit_results['sigma'], loc=fit_results['loc'], scale=np.exp(fit_results['mu']))


def get_normal_dist(fit_results):
    return stat.norm(scale=fit_results['scale'], loc=fit_results['loc'])


def get_negbinomial_dist(fit_results):
    return stat.nbinom(n=fit_results['n'], p=fit_results['p'], loc=fit_results['loc'])


def get_poisson_dist(fit_results):
    return stat.poisson(mu=fit_results['mu'], loc=fit_results['loc'])


def get_triangular_dist(fit_results):
    return stat.triang(c=fit_results['c'], scale=fit_results['scale'], loc=fit_results['loc'])


def get_uniform_dist(fit_results):
    return stat.uniform(scale=fit_results['scale'], loc=fit_results['loc'])


def get_uniform_discrete_dist(fit_results):
    return stat.randint(low=int(fit_results['l']), high=int(fit_results['u'])+1)


def get_weibull_dist(fit_results):
    return stat.weibull_min(c=fit_results['c'], scale=fit_results['scale'], loc=fit_results['loc'])


# ---- plot the fit of different distributions ----
def plot_beta_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                  fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "a", "b", "loc", "scale"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_beta_dist(fit_results),
        label='Beta',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_beta_binomial_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                           fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "a", "b", "n", and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_discrete(
        data=data,
        dist=get_beta_binomial_dist(fit_results),
        label='Beta-Binomial',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_binomial_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                      fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "n", "p", and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_discrete(
        data=data,
        dist=get_binomial_dist(fit_results),
        label='Binomial',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_exponential_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                         fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "scale" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_exponential_dist(fit_results),
        label='Exponential',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_gamma_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                   fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "a", "scale" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_gamma_dist(fit_results),
        label='Gamma',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename
    )


def plot_gamma_poisson_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                           fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "a", "gamma_scale", "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_discrete(
        data=data,
        dist=get_gamma_poisson_dist(fit_results),
        label='Gamma-Poisson',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_geometric_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                       fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "p" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_discrete(
        data=data,
        dist=get_geometric_dist(fit_results),
        label='Geometric',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_johnson_sb_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                        fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "a", "b", "loc", "scale", and "AIC"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_johnson_sb_dist(fit_results),
        label='Johnson Sb',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_johnson_su_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                        fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "a", "b", "loc", "scale", and "AIC"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_johnson_su_dist(fit_results),
        label='Johnson Su',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_lognormal_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                       fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "mu", "sigma" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_lognormal_dist(fit_results),
        label='LogNormal',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename
    )


def plot_normal_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                    fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "loc" and "scale"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_normal_dist(fit_results),
        label='Normal',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename
    )


def plot_negbinomial_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                         fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "n", "p" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_discrete(
        data=data,
        dist=get_negbinomial_dist(fit_results),
        label='Negative Binomial',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_poisson_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                     fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "mu" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_discrete(
        data=data,
        dist=get_poisson_dist(fit_results),
        label='Poisson',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_triangular_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                        fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "c", "loc", "scale",
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_triangular_dist(fit_results),
        label='Triangular',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename
    )


def plot_uniform_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                     fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "loc", "scale",
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_uniform_dist(fit_results),
        label='Uniform',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename
    )


def plot_uniform_discrete_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                              fig_size=(6, 5), bin_width=1, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "mu" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_discrete(
        data=data,
        dist=get_uniform_discrete_dist(fit_results),
        label='Uniform-Discrete',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename)


def plot_weibull_fit(data, fit_results, title=None, x_label=None, x_range=None, y_range=None,
                     fig_size=(6, 5), bin_width=None, filename=None):
    """
    :param data: (numpy.array) observations
    :param fit_results: dictionary with keys "c", "scale" and "loc"
    :param title: title of the figure
    :param x_label: label to show on the x-axis of the histogram
    :param x_range: (tuple) x range
    :param y_range: (tuple) y range
        (the histogram shows the probability density so the upper value of y_range should be 1).
    :param fig_size: int, specify the figure size
    :param bin_width: bin width
    :param filename: filename to save the figure as
    """

    plot_fit_continuous(
        data=data,
        dist=get_weibull_dist(fit_results),
        label='Weibull',
        bin_width=bin_width, title=title, x_label=x_label, x_range=x_range, y_range=y_range,
        fig_size=fig_size, filename=filename
    )





import math


class Line:

    def __init__(self, x1, x2, y1, y2):

        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2

        self.slope = (y2 - y1)/(x2 - x1)
        self.intercept = self.get_y(x=0)

    def get_y(self, x):

        return self.slope * (x - self.x1) + self.y1

    def get_intercept_with_x_axis(self):
        
        try:
            value = - self.y1/self.slope + self.x1
        except ValueError:
            value = math.nan
        
        return value


if __name__ == '__main__':
    test = Line(x1=1, x2=2, y1=0, y2=2)
    print(test.get_intercept_with_x_axis())



import matplotlib.pyplot as plt
import numpy as np


def plot_pyramids(observed_data, simulated_data=None,
                  title=None, fig_size=None,
                  x_lim=100,
                  y_labels=None,
                  age_group_width=5,
                  colors=('brown', 'grey', 'blue'),
                  legend=('Data', 'Model'),
                  length_of_sim_bars=500,
                  linewidths_of_sim_bars=1,
                  scale_of_sim_legend=0.5,
                  transparency=0.5,
                  file_name=None):
    """
    :param observed_data: example:
                        [[0, 0, 0.1], [0, 1, 0.2], [5, 0, 0.3], [5, 1, 0.4], [10, 0, 0.6], [10, 1, 0.4]]
    :param simulated_data: example:
                        [
                            [[0, 0, 0.5], [0, 1, 0.3], [5, 0, 0.5], [5, 1, 0.6]],
                            [[0, 0, 0.2], [0, 1, 0.4], [5, 0, 0.9], [5, 1, 0.4]]
                        ]
    :param title: (string) figure title
    :param fig_size: (tuple) figure size (e.g. (2, 4))
    :param x_lim: maximum of x-axis (between 0 and 100)
    :param y_labels: (list) of strings for labels of y_axis
    :param age_group_width: width of each age group in years.
    :param colors: (tuple) example: ('brown', 'grey', 'blue')
                                    for men, women and simulation bars
    :param legend: (tuple) default ('Data', 'Model').
                    legend will not be displayed if set to None
    :param length_of_sim_bars: length of simulation bars
    :param linewidths_of_sim_bars: line width of the simulation bars
    :param scale_of_sim_legend: (between 0 and 1) to shrink simulation bars
                                shown on the legends
    :param transparency: transparency of bars
    :param file_name: (string) to save the figure. Example: 'Pyramid.png'.
                    if not provided, the figure will be displayed.
    """

    w_sizes = []    # sizes of female age groups
    m_sizes = []    # sizes of male age groups
    w_ages = []     # age breaks for women
    m_ages = []     # age breaks for men

    # number of sex and age groups
    num_of_groups = len(observed_data)

    i = 0
    # sort data into separate arrays
    while i < num_of_groups:
        # woman
        if observed_data[i][1] == 1:
            w_sizes.append(observed_data[i][2] * 100)
            w_ages.append(observed_data[i][0])
        # men
        else:
            m_sizes.append(observed_data[i][2] * 100)
            m_ages.append(observed_data[i][0])
        i = i + 1

    if m_ages != w_ages:
        raise ValueError('Male and female age groups should be the same.')

    # find maximum value of the y_axis
    y_lim = m_ages[-1] - m_ages[0] + age_group_width / 2

    # either sets to default ticks or user input
    y_ticks_places = range(0, len(m_ages) * age_group_width, age_group_width)

    # either sets to default tick labels or user input
    if y_labels is None:
        y_labels = ['0-4', '5-9', '10-14', '15-19', '20-24', '25-29', '30-34', '35-39', '40-44',
                    '45-49', '50-54', '55-59', '60-64', '65-69', '70-74', '75-79',
                    '80-84', '85-89', '90-94', '95-100', '100+']

    # plot configuration
    fig, axis = plt.subplots(ncols=2, sharey=True, tight_layout=True, figsize=fig_size)
    # figure title
    st = fig.suptitle(title)
    # add y labels and ticks
    plt.setp(axis, yticks=y_ticks_places, yticklabels=y_labels[:len(y_ticks_places)])

    # ranges of x- and y-axes
    for i in [0, 1]:
        axis[i].xmin, axis[i].xmax, axis[i].ymin, axis[i].ymax \
            = axis[i].axis([0, x_lim, -age_group_width * (1 / 2 + 1 / 10), y_lim + age_group_width * 1 / 10])

    # bar plot for males
    axis[0].barh(np.array(m_ages)-m_ages[0], m_sizes, age_group_width,
                 align='center', color=colors[0],
                 edgecolor='black', alpha=transparency, label='Data')
    axis[0].invert_xaxis()

    # bar plot for females
    axis[1].barh(np.array(w_ages)-w_ages[0], w_sizes, age_group_width,
                 align='center', color=colors[1],
                 edgecolor='black', alpha=transparency, label='Data')

    # title of each sub-figure
    axis[0].set(title='\nMen')
    axis[1].set(title='\nWomen')

    # add labels
    axis[0].set_ylabel('Age')
    axis[0].set_xlabel('Percent of Population')
    axis[1].set_xlabel('Percent of Population')

    # adding pyramids from simulation
    if simulated_data is not None:
        n_replications = len(simulated_data)
        rep = 0     # simulation replication
        g = 0       # age-sex group
        sim_data_w = []
        sim_data_m = []
        sim_data_wa = []
        sim_data_ma = []
        while rep < n_replications:
            while g < len(simulated_data[rep]):
                if simulated_data[rep][g][1] == 1:
                    sim_data_w.append(simulated_data[rep][g][2] * 100)
                    sim_data_wa.append(simulated_data[rep][g][0] - w_ages[0])
                    g = g + 1
                else:
                    sim_data_m.append(simulated_data[rep][g][2] * 100)
                    sim_data_ma.append(simulated_data[rep][g][0] - m_ages[0])
                    g = g + 1
            g = 0
            rep = rep + 1
        axis[0].scatter(sim_data_m, sim_data_ma,
                        marker='|', linewidths=linewidths_of_sim_bars,
                        s=length_of_sim_bars,
                        color=colors[2], alpha=transparency,
                        label='Model')
        axis[1].scatter(sim_data_w, sim_data_wa,
                        marker='|', linewidths=linewidths_of_sim_bars,
                        s=length_of_sim_bars,
                        color=colors[2], alpha=transparency,
                        label='Model')
    else:
        pass

    if legend is not None:
        handles, labels = axis[1].get_legend_handles_labels()
        handlesa, labelsa = axis[0].get_legend_handles_labels()
        axis[0].legend(reversed(handlesa), reversed(labelsa),
                       markerscale=scale_of_sim_legend)
        axis[1].legend(reversed(handles), reversed(labels),
                       markerscale=scale_of_sim_legend)

    fig.tight_layout()

    if file_name is None:
        fig.show()
    else:
        fig.savefig(file_name, dpi=300)



from math import pow, exp

import matplotlib.pyplot as plt

from deampy.plots.plot_support import output_figure


class _StepSizeRule:

    def get_epsilon(self, itr):
        # goes to zero as itr goes to infinity
        pass

    def get_learning_rate(self, itr):
        # goes to 1 as itr goes to infinity

        return 1 - self.get_epsilon(itr)


class EpsilonGreedy(_StepSizeRule):

    def __init__(self, formula, beta, min=0, max=1):
        """
            For selecting the greedy action with probability 1-epsilon.
            for pow decay formula: epsilon_n = min + (max-min)/n^beta, beta over (0.5, 1], n > 0
            for exp decay formula: epsilon_n = min + (max-min) * exp(-beta * n), beta > 0, n > 0
        :param formula: (str) 'pow' or 'exp'
        :param beta: (float) the decay rate
        :param min: (float) the minimum epsilon
        :param max: (float) the maximum epsilon
        """

        if formula not in ['pow', 'exp']:
            raise ValueError('Invalid formula. Choose between "pow" and "exp".')
        if beta <= 0:
            raise ValueError('Beta should be greater than 0.')
        if min < 0 or max > 1:
            raise ValueError('Epsilon should be between 0 and 1.')

        self._formula = formula
        self._beta = beta
        self._min = min
        self._max = max

    def __str__(self):
        """
        :return: (str) the formula and the parameters
        """
        return '{}-beta{}-max{}-min{}'.format(self._formula, self._beta, self._max, self._min)

    def get_epsilon(self, itr):

        if itr <= 0:
            return self._max

        if self._formula == 'pow':
            return self._min + (self._max - self._min) * pow(itr, -self._beta)
        elif self._formula == 'exp':
            return self._min + (self._max - self._min) * exp(-self._beta * itr)
        else:
            raise ValueError('Invalid formula. Choose between "pow" and "exp".')

    @staticmethod
    def plot(formula, betas, maxs, mins, n_itrs, fig_size=None, fig_filename=None):
        """
        plots the exploration and learning rates
        :param formula: (str) 'pow' or 'exp'
        :param betas: (list) of decay rates
        :param maxs: (list) of maximum epsilon values
        :param mins: (list) of minimum epsilon values
        :param n_itrs: (int) number of iterations
        :param fig_size: (tuple) the size of the figure
        :param fig_filename: (str) the filename to save the figure
        """

        x = range(1, n_itrs + 1)

        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=fig_size)

        for max in maxs:
            for min in mins:
                for beta in betas:
                    rule = EpsilonGreedy(formula=formula, beta=beta, max=max, min=min)
                    y = [rule.get_epsilon(i) for i in x]
                    ax[0].plot(x, y, label=str(rule))
                    y = [rule.get_learning_rate(i) for i in x]
                    ax[1].plot(x, y, label=str(rule))

        ax[0].axhline(y=0, color='black', linestyle='--')
        ax[1].axhline(y=1, color='black', linestyle='--')

        ax[0].set_xlabel('Iteration')
        ax[1].set_xlabel('Iteration')

        ax[0].set_ylabel('Exploration Rate')
        ax[1].set_ylabel('Learning Rate')

        # ax.set_title('Epsilon-Greedy Exploration Rule')
        ax[0].legend()
        ax[1].legend()
        fig.tight_layout()
        if fig_filename:
            output_figure(fig, fig_filename)
        else:
            fig.show()


class Harmonic(_StepSizeRule):
    # step_n = b / (b + n - 1), for n > 0 and b >= 1
    # (i is the iteration of the optimization algorithm)

    def __init__(self, b):
        self._b = b

    def __str__(self):
        return 'b{}'.format(self._b)

    def get_epsilon(self, itr):
        if itr <= 0:
            return 1
        return self._b / (self._b + itr - 1)

    @staticmethod
    def plot(bs, n_itrs):

        x = range(1, n_itrs + 1)

        fig, ax = plt.subplots()
        for b in bs:
            rule = Harmonic(b)
            y = [rule.get_learning_rate(i) for i in x]
            ax.plot(x, y, label=str(rule))
        ax.plot(x, y)
        ax.axhline(y=1, color='black', linestyle='--')
        ax.set_xlabel('Iteration')
        ax.set_ylabel('Forgetting Factor')
        ax.set_title('Harmonic Learning Rule')
        ax.legend()
        fig.show()


if __name__ == '__main__':

    EpsilonGreedy.plot(
        formula='exp', maxs=[0.5], mins=[0.025], betas=[0.01, 0.02, .03], n_itrs=1000)
    EpsilonGreedy.plot(
        formula='pow', maxs=[0.5], mins=[0.025], betas=[0.5, 0.7, 0.9], n_itrs=1000)

    # Harmonic.plot(bs=[1, 10, 20], n_itrs=1000)



import math
import warnings

import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
import scipy.stats as stat
from numpy.random import RandomState
from scipy.optimize import fmin_slsqp
from scipy.optimize import minimize

warnings.filterwarnings("ignore")


def AIC(k, log_likelihood):
    """ :returns Akaike information criterion"""
    return 2 * k - 2 * log_likelihood


class RNG(RandomState):
    def __init__(self, seed):
        RandomState.__init__(self, seed)

    def sample(self):
        return self.random_sample()


class RVG:
    def __init__(self):
        pass

    def sample(self, rng, arg=None):
        """
        :param rng: an instant of RNG class
        :param arg: optional arguments
        :returns one realization from the defined probability distribution """

        # abstract method to be overridden in derived classes to process an event
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_mean_st_dev(self):
        """ :returns: (tuple) of mean and standard deviation """
        pass

    def get_percentile_interval(self, alpha=0.05):
        """ :returns: [l, u], where l and u are the lower and upper critical values
                of the distribution"""

    @staticmethod
    def _get_percentile_interval(alpha, dist, params):
        """
        :param alpha: confidence level
        :param dist: a distribution with .ppf method
        :param params: parameters of the distribution
        :return:
        """
        return [dist.ppf(alpha / 2, *params),
                dist.ppf(1 - alpha / 2, *params)]


class Constant (RVG):
    def __init__(self, value):
        RVG.__init__(self)
        self.value = value

    def sample(self, rng, arg=None):
        return self.value


class Bernoulli(RVG):
    def __init__(self, p):
        """
        E[X] = p
        Var[X] = p(1-p)
        """
        RVG.__init__(self)
        self.p = p

    def sample(self, rng, arg=None):
        sample = 0
        if rng.random_sample() <= self.p:
            sample = 1
        return sample


class Beta(RVG):
    def __init__(self, a, b, loc=0, scale=1):
        """
        E[X] = a/(a + b)*scale + loc
        Var[X] = (scale**2) ab/[(a + b)**2(a + b + 1)]
        min[X] = loc
        max[x] = min[X] + scale
        """
        RVG.__init__(self)
        self.a = a
        self.b = b
        self.scale = scale
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.beta(self.a, self.b) * self.scale + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.beta,
            params=[self.a, self.b, self.loc, self.scale])

    @staticmethod
    def fit_mm(mean, st_dev, minimum=0, maximum=1):
        """
        :param mean: sample mean
        :param st_dev: sample standard deviation
        :param minimum: fixed minimum
        :param maximum: fixed maximum
        :return: dictionary with keys "a", "b", "loc" and "scale"
        """

        if not (minimum < mean < maximum):
            raise ValueError('Mean should be between minimum {} and maximum {}.'.format(minimum, maximum))

        # shift the distribution by loc and scale
        mean = (mean - minimum) * 1.0 / (maximum - minimum)
        st_dev = st_dev * 1.0 / (maximum - minimum)

        a_plus_b = mean * (1 - mean) / pow(st_dev, 2) - 1
        a = mean * a_plus_b

        return {"a": a, "b": a_plus_b - a, "loc": minimum, "scale": maximum - minimum}

    @staticmethod
    def fit_ml(data, minimum=None, maximum=None):
        """
        :param data: (numpy.array) observations
        :param minimum: minimum of data (calculated from data if not provided)
        :param maximum: maximum of data (calculated from data if not provided)
        :returns: dictionary with keys "a", "b", "loc", "scale", and "AIC"
        """

        # transform data into [0,1]
        L = np.min(data) if minimum is None else minimum
        U = np.max(data) if maximum is None else maximum
        data = (data - L) / (U - L)

        # estimate the parameters
        a, b, loc, scale = stat.beta.fit(data, floc=0)

        # calculate AIC
        aic = AIC(k=3,
                  log_likelihood=np.sum(stat.beta.logpdf(data, a, b, loc, scale)))

        # report results in the form of a dictionary
        return {"a": a, "b": b, "loc": L, "scale": U - L, "AIC": aic}

    @staticmethod
    def get_uncertainty_interval(alpha, mean, st_dev, minimum=0, maximum=1):
        """
        :param alpha: significance level (use 5% to get 95% uncertainty interval)
        :param mean: sample mean
        :param st_dev: sample standard deviation
        :param minimum: fixed minimum
        :param maximum: fixed maximum
        :return: (list) with specified percentiles
        """

        beta_par = Beta.fit_mm(mean=mean, st_dev=st_dev, minimum=minimum, maximum=maximum)

        l = stat.beta.ppf(q=alpha/2,
                          a=beta_par['a'], b=beta_par['b'], loc=beta_par['loc'], scale=beta_par['scale'])
        u = stat.beta.ppf(q=1-alpha/2,
                          a=beta_par['a'], b=beta_par['b'], loc=beta_par['loc'], scale=beta_par['scale'])

        return [l, u]


class BetaBinomial(RVG):
    def __init__(self, n, a, b, loc=0):
        """
        E[X] = na/(a+b) + loc
        Var[X] = (nab(a+b+n))/((a+b)**2(a+b+1))
        """
        RVG.__init__(self)
        self.n = n
        self.a = a
        self.b = b
        self.loc = loc

    def sample(self, rng, arg=None):
        """
        ref: https://blogs.sas.com/content/iml/2017/11/20/simulate-beta-binomial-sas.html
        :return: a realization from the Beta Binomial distribution
        """
        sample_p = rng.beta(self.a, self.b)
        sample_n = rng.binomial(self.n, sample_p)

        return sample_n + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.betabinom,
            params=[self.n, self.a, self.b, self.loc])

    @staticmethod
    def fit_mm(mean, st_dev, n, fixed_location=0):
        """
        # ref: https://en.wikipedia.org/wiki/Beta-binomial_distribution
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :param n: the number of trials in the Binomial distribution
        :param fixed_location: location, 0 by default
        :return: dictionary with keys "a", "b", "n", "loc", and "scale"
        """
        mean = 1.0 * (mean - fixed_location)
        variance = st_dev ** 2.0
        m2 = variance + mean ** 2  # second moment

        a1 = n * mean - m2
        a2 = n * (m2 / mean - mean - 1) + mean
        a = a1 / a2
        b1 = (n - mean) * (n - m2 / mean)
        b2 = a2
        b = b1 / b2

        return {"a": a, "b": b, "n": n, "loc": fixed_location}

    @staticmethod
    def ln_pmf(a, b, n, k):
        part_1 = sp.special.comb(n, k)
        part_2 = sp.special.betaln(k + a, n - k + b)
        part_3 = sp.special.betaln(a, b)
        result = (np.log(part_1) + part_2) - part_3
        return result

    # define log_likelihood function: sum of log(pmf) for each data point
    @staticmethod
    def ln_likelihood(a, b, n, data):
        n = int(np.round(n, 0))
        result = 0
        for i in range(len(data)):
            result += BetaBinomial.ln_pmf(a=a, b=b, n=n, k=data[i])
        return result

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: fixed location
        :returns: dictionary with keys "a", "b", "n" and "AIC"
        """

        data = 1.0 * (data - fixed_location)

        def neg_ln_l(a_b_n, data):
            return -BetaBinomial.ln_likelihood(a=a_b_n[0], b=a_b_n[1], n=a_b_n[2], data=data)

        # estimate the parameters by minimize negative log-likelihood
        # initialize parameters
        initial_guess = [1, 1, np.max(data)]
        # call Scipy optimizer to minimize the target function
        # with bounds for a [0, 10], b [0, 10] and n [0,100+max(data)]
        result = minimize(
            neg_ln_l,
            x0=np.array(initial_guess),
            args=(np.asarray(data),),
            method='L-BFGS-B',
            bounds=[(1e-6, None), (1e-6, None), (0, None)]
        )

        if not result.success:
            raise RuntimeError("Optimization failed: " + result.message)

        a, b, n = result.x


        # fitted_a_b_n, value, iter, imode, smode \
        #     = fmin_slsqp(neg_ln_l, a_b_n_0,
        #                  bounds=[(0.0, 10000.0), (0.0, 10000.0), (0, np.max(data) + 100)],
        #                  disp=False, full_output=True)

        # calculate AIC
        aic = AIC(
            k=3,
            log_likelihood=BetaBinomial.ln_likelihood(
                a=a, b=b, n=int(n), data=data)
        )

        # report results in the form of a dictionary
        return {"a": a, "b": b, "n": int(n), "loc": fixed_location, "AIC": aic}

    @staticmethod
    def fit_ml_old(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: fixed location
        :returns: dictionary with keys "a", "b", "n" and "AIC"
        """

        data = 1.0 * (data - fixed_location)

        def neg_ln_l(a_b_n):
            return -BetaBinomial.ln_likelihood(a=a_b_n[0], b=a_b_n[1], n=a_b_n[2], data=data)

        # estimate the parameters by minimize negative log-likelihood
        # initialize parameters
        a_b_n_0 = [1, 1, np.max(data)]
        # call Scipy optimizer to minimize the target function
        # with bounds for a [0, 10], b [0, 10] and n [0,100+max(data)]
        fitted_a_b_n, value, iter, imode, smode \
            = fmin_slsqp(neg_ln_l, a_b_n_0,
                         bounds=[(0.0, 10000.0), (0.0, 10000.0), (0, np.max(data) + 100)],
                         disp=False, full_output=True)

        # calculate AIC
        aic = AIC(
            k=3,
            log_likelihood=BetaBinomial.ln_likelihood(
                a=fitted_a_b_n[0], b=fitted_a_b_n[1], n=fitted_a_b_n[2], data=data)
        )

        # report results in the form of a dictionary
        return {"a": fitted_a_b_n[0], "b": fitted_a_b_n[1], "n": fitted_a_b_n[2], "loc": fixed_location, "AIC": aic}


class Binomial(RVG):
    def __init__(self, n, p, loc=0):
        """
        E[X] = np + loc
        Var[X] = np(1-p)
        """
        RVG.__init__(self)
        self.n = n
        self.p = p
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.binomial(self.n, self.p) + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.binom,
            params=[self.n, self.p, self.loc])

    @staticmethod
    def fit_mm(mean, st_dev, fixed_location=0):
        """
        :param mean: sample mean
        :param st_dev: sample standard deviation
        :param fixed_location: fixed location, 0 by default
        :return: dictionary with keys "p", "n" and "loc"
        """
        mean = mean - fixed_location
        p = 1.0 - (st_dev ** 2) / mean
        n = mean / p

        return {"n": n, "p": p, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: fixed location
        :returns: dictionary with keys "p", "n", "loc" and "AIC"
        """

        # the MLE of p is x/n
        # if we have N data point with Xi~Bin(n,p), then sum(Xi)~Bin(n*N,p), p_hat = sum(xi)/(n*N)
        # # https://onlinecourses.science.psu.edu/stat504/node/28

        data = data - fixed_location

        mean = np.mean(data)
        st_dev = np.std(data)
        p = 1.0 - (st_dev ** 2) / mean
        n = mean / p

        # calculate AIC
        aic = AIC(
            k=1,
            log_likelihood=np.sum(stat.binom.logpmf(data, n, p))
        )

        # report results in the form of a dictionary
        return {"n": n, "p": p, "loc": fixed_location, "AIC": aic}


class Dirichlet(RVG):
    def __init__(self, a, if_ignore_0s=False):
        """
        E[Xi] = ai/a0 where a0 = sum of ai's.
        Var[Xi] = (ai(a0-ai))/((a0)**2(a0+1)) where a0 = sum of ai's.
        :param a: array or list
        :param if_ignore_0s: (bool) numpy requires all elements of 'a' to be >0. Setting this
            parameter to True allows 'a' to contain 0s.
        """
        RVG.__init__(self)
        self.a = a
        self.ifIgnore0s = if_ignore_0s
        self.nonZeroA = []
        self.idxOfNonZeroA = []
        for i, value in enumerate(a):
            if value > 0:
                self.nonZeroA.append(value)
                self.idxOfNonZeroA.append(i)

    def sample(self, rng, arg=None):
        """
        :return: (array) a realization from the Dirichlet distribution
        """
        if self.ifIgnore0s:
            sample = rng.dirichlet(self.nonZeroA)
            result = [0]*len(self.a)
            for i, idx in enumerate(self.idxOfNonZeroA):
                result[idx] = sample[i]
            return result
        else:
            return rng.dirichlet(self.a)


class Empirical(RVG):
    def __init__(self, probabilities):
        """
        assuming outcomes = [0, 1, 2, 3, ...]
        E[X] = sum(outcome*prob)
        Var[X] = sum((outcome**2)*prob) - E[X]**2
        """
        RVG.__init__(self)

        self.prob = np.array(probabilities)
        self.nOutcomes = len(self.prob)

        if self.prob.sum() < 0.99999 or self.prob.sum() > 1.00001:
            raise ValueError('Probabilities should sum to 1.')
        self.prob = probabilities

    def sample(self, rng, arg=None):
        """
        :return: (int) from possible outcomes [0, 1, 2, 3, ...]
        """
        # this works for both numpy array and list
        # ref:https://stackoverflow.com/questions/4265988/generate-random-numbers-with-a-given-numerical-distribution
        return rng.choice(range(self.nOutcomes), p=self.prob)

    @staticmethod
    def fit_mm(data, bin_size=1):
        """
        :param data: (numpy.array) observations
        :param bin_size: float, the width of histogram's bins
        :returns: dictionary keys of "bins" and "freq"
        """

        l = np.min(data)
        u = np.max(data) + bin_size
        n_bins = math.floor((u - l)/bin_size) + 1
        bins = np.linspace(l, u, n_bins)

        result = plt.hist(data, bins=bins)

        bins = result[1]  # bins are in the form of [a,b)
        freq = result[0] * 1.0 / len(data)

        return {"bins": bins, "freq": freq}


class Exponential(RVG):
    def __init__(self, scale, loc=0):
        """
        E[X] = scale + loc
        Var[X] = scale**2
        """
        RVG.__init__(self)
        self.scale = scale
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.exponential(scale=self.scale) + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.expon,
            params=[self.loc, self.scale])

    @staticmethod
    def fit_mm(mean, fixed_location=0):
        """
        :param mean: sample mean
        :param fixed_location: minimum of the exponential distribution, set to 0 by default
        :return: dictionary with keys "loc" and "scale"
        """
        mean = mean - fixed_location
        scale = mean

        return {"scale": scale, "loc": fixed_location}

    @ staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: fixed location
        :returns: dictionary with keys "loc", "scale", and "AIC"
        """
        # estimate the parameters of exponential
        loc, scale = stat.expon.fit(data, floc=fixed_location)

        # calculate AIC
        aic = AIC(
            k=1,
            log_likelihood=np.sum(stat.expon.logpdf(data, loc, scale)))

        # report results in the form of a dictionary
        return {"scale": scale, "loc": loc, "AIC": aic}


class Gamma(RVG):
    def __init__(self, shape, scale=1, loc=0):
        """
        E[X] = shape*scale + loc
        Var[X] = shape*scale**2
        """
        RVG.__init__(self)
        self.shape = shape
        self.scale = scale
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.gamma(self.shape, self.scale) + self.loc

    def get_mean_st_dev(self):
        return self.shape * self.scale + self.loc, math.sqrt(self.shape * self.scale ** 2)

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(alpha=alpha, dist=stat.gamma,
                                             params=[self.shape, self.loc, self.scale])

    @staticmethod
    def fit_mm(mean, st_dev, fixed_location=0):
        """
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :param fixed_location: location, 0 by default
        :return: dictionary with keys "shape", "scale", and "loc"
        """
        mean = mean - fixed_location

        shape = (mean / st_dev) ** 2
        scale = (st_dev ** 2) / mean

        # report results in the form of a dictionary
        return {"shape": shape, "scale": scale, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: fixed location
        :returns: dictionary with keys "shape", "scale", "loc", and "AIC"
        """

        # estimate the parameters of gamma
        # alpha = a, beta = 1/scale
        a, loc, scale = stat.gamma.fit(data, floc=fixed_location)

        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.gamma.logpdf(data, a, loc, scale)))

        # report results in the form of a dictionary
        return {"shape": a, "scale": scale, "loc": loc, "AIC": aic}


class GammaPoisson(RVG):
    #
    def __init__(self, shape, scale, loc=0):
        """
        E[X] = (shape*scale) + loc
        Var[X] = shape*scale + shape*(scale**2)
        """
        RVG.__init__(self)
        self.shape = shape
        self.scale = scale
        self.loc = loc

    def sample(self, rng, arg=None):
        sample_rate = Gamma(shape=self.shape, scale=self.scale).sample(rng)
        sample_poisson = Poisson(mu=sample_rate)
        return sample_poisson.sample(rng) + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return [self.ppf(q=alpha/2), self.ppf(q=1-alpha/2)]

    def pmf(self, k):
        # https: // en.wikipedia.org / wiki / Negative_binomial_distribution  # Gamma%E2%80%93Poisson_mixture
        # with r= shape and p = scale/(scale + 1)

        if type(k) == int:
            k = k - self.loc
            p = self.scale / (self.scale + 1)
            part1 = 1.0 * sp.special.gamma(self.shape + k) / (sp.special.gamma(self.shape) * sp.special.factorial(k))
            part2 = (p ** k) * ((1 - p) ** self.shape)

            return part1 * part2
        else:
            result = []
            for this_k in k:
                result.append(self.pmf(this_k))
            return result

    def ppf(self, q):
        """ :returns: percentage point function (inverse of cdf)"""

        cum = 0
        k = self.loc

        while True:
            cum += self.pmf(k)
            k += 1
            if cum > q:
                break
        return k - 1

    @staticmethod
    def ln_pmf(a, gamma_scale, loc, k):
        # https: // en.wikipedia.org / wiki / Negative_binomial_distribution  # Gamma%E2%80%93Poisson_mixture
        # with r= a and p = scale/(scale + 1)

        k = k - loc
        p = gamma_scale / (gamma_scale + 1)
        part1 = np.log(sp.special.gamma(a + k)) \
                - np.log(sp.special.gamma(a)) \
                - np.log(sp.special.factorial(k))
        part2 = k * np.log(p) + a * np.log(1 - p)

        return part1 + part2

    # define log_likelihood function: sum of log(pmf) for each data point
    @staticmethod
    def ln_likelihood(a, scale, loc, data):
        result = 0
        for k in data:
            v = GammaPoisson.ln_pmf(a=a, gamma_scale=scale, loc=loc, k=k)
            result += v
        return result

    @staticmethod
    def fit_mm(mean, st_dev, fixed_location=0):
        """
        :param mean: sample mean
        :param st_dev: sample standard deviation
        :param fixed_location: location, 0 by default
        :return: dictionary with keys "shape", "scale", "loc" and "scale"
        """
        # ref: http://www.math.wm.edu/~leemis/chart/UDR/PDFs/Gammapoisson.pdf
        # scale = 1/beta

        mean = mean - fixed_location
        variance = st_dev**2.0

        scale = (variance-mean)/mean
        shape = mean/scale

        return {"shape": shape, "scale": scale, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):

        def neg_log_likelihood(params, data):
            return -GammaPoisson.ln_likelihood(a=params[0], scale=params[1], loc=fixed_location, data=data)

        initial_guess = (1.0, 1.0)

        result = minimize(
            neg_log_likelihood,
            x0=np.array(initial_guess),
            args=(np.asarray(data),),
            method='L-BFGS-B',
            bounds=[(1e-6, None), (1e-6, None)]
        )

        if not result.success:
            raise RuntimeError("Optimization failed: " + result.message)

        r_hat, p_hat = result.x

        # calculate AIC (note that the data has already been shifted by loc)
        aic = AIC(
            k=2,
            log_likelihood=GammaPoisson.ln_likelihood(
                a=r_hat, scale=p_hat, loc=fixed_location, data=data)
        )

        return {"shape": r_hat, "scale": p_hat, "loc": fixed_location, "AIC": aic}

    @staticmethod
    def fit_ml_old(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location
        :return: dictionary with keys "shape", "scale", "loc", and "AIC"
        """

        data = data - fixed_location

        # define negative log-likelihood, the target function to minimize
        def neg_ln_l(shape_scale, data):
            return -GammaPoisson.ln_likelihood(a=shape_scale[0], scale=shape_scale[1], loc=0, data=data)

        # estimate the parameters by minimize negative log-likelihood
        # initialize parameters
        initial_guess = [1, 1]
        # call Scipy optimizer to minimize the target function
        # with bounds for p [0,1] and r [0,10]
        fitted_shape_scale, value, iter, imode, smode = fmin_slsqp(neg_ln_l, initial_guess,
                                                               #bounds=[(0.0, 10.0), (0, 1)],
                                                               disp=False, full_output=True)

        # calculate AIC (note that the data has already been shifted by loc)
        aic = AIC(
            k=2,
            log_likelihood=GammaPoisson.ln_likelihood(
                a=fitted_shape_scale[0], scale=fitted_shape_scale[1], loc=0, data=data)
        )
        return {"shape": fitted_shape_scale[0], "scale": fitted_shape_scale[1], "loc": fixed_location, "AIC": aic}


class Geometric(RVG):
    def __init__(self, p, loc=0):
        """
        E[X] = 1/p+loc
        Var[X] = (1-p)/p**2
        """
        RVG.__init__(self)
        self.p = p
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.geometric(self.p) + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.geom,
            params=[self.p, self.loc])

    @staticmethod
    def fit_mm(mean, fixed_location=0):
        """
        :param mean: sample mean
        :param fixed_location: location, 0 by default
        :return: dictionary with keys "p", "loc"
        """
        mean = mean - fixed_location
        p = 1.0/mean

        return {"p": p, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :return: dictionary with keys "p", "loc"
        """

        # https://www.projectrhea.org/rhea/index.php/MLE_Examples:_Exponential_and_Geometric_Distributions_Old_Kiwi
        data = data - fixed_location
        p = len(data) * 1.0 / np.sum(data)

        # calculate AIC
        aic = AIC(
            k=1,
            log_likelihood=np.sum(stat.geom.logpmf(data, p)))

        # report results in the form of a dictionary
        return {"p": p, "loc": fixed_location, "AIC": aic}


class JohnsonSb(RVG):
    def __init__(self, a, b, loc, scale):
        """
        The moments of the Johnson SB distribution do not have a simple expression.
        E[X] = theoretical value give by SciPy johnsonsb.mean(a,b,loc,scale)
        Var[X] = theoretical value give by SciPy johnsonsb.var(a,b,loc,scale)
        """
        RVG.__init__(self)
        self.a = a
        self.b = b
        self.loc = loc
        self.scale = scale

    def sample(self, rng, arg=None):
        return stat.johnsonsb.rvs(self.a, self.b, self.loc, self.scale, random_state=rng)

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.johnsonsb,
            params=[self.a, self.b, self.loc, self.scale])

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "a", "b", "loc", "scale", and "AIC"
        """

        # estimate the parameters
        a, b, loc, scale = stat.johnsonsb.fit(data, floc=fixed_location)

        # calculate AIC
        aic = AIC(
            k=3,
            log_likelihood=np.sum(stat.johnsonsb.logpdf(data, a, b, loc, scale)))

        # report results in the form of a dictionary
        return {"a": a, "b": b, "loc": loc, "scale": scale, "AIC": aic}


class JohnsonSu(RVG):
    def __init__(self, a, b, loc, scale):
        """
        The moments of the Johnson SU distribution do not have a simple expression.
        E[X] = theoretical value give by SciPy johnsonsu.mean(a,b,loc,scale)
        Var[X] = theoretical value give by SciPy johnsonsu.var(a,b,loc,scale)
        """
        RVG.__init__(self)
        self.a = a
        self.b = b
        self.loc = loc
        self.scale = scale

    def sample(self, rng, arg=None):
        return stat.johnsonsu.rvs(a=self.a, b=self.b, loc=self.loc, scale=self.scale, random_state=rng)

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.johnsonsu,
            params=[self.a, self.b, self.loc, self.scale])

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "a", "b", "loc", "scale", and "AIC"
        """

        # estimate the parameters
        a, b, loc, scale = stat.johnsonsu.fit(data, floc=fixed_location)

        # calculate AIC
        aic = AIC(
            k=3,
            log_likelihood=np.sum(stat.johnsonsu.logpdf(data, a, b, loc, scale)))

        # report results in the form of a dictionary
        return {"a": a, "b": b, "loc": loc, "scale": scale, "AIC": aic}


class LogNormal(RVG):
    def __init__(self, mu, sigma, loc=0):
        """
        E[X] = exp(mu + 1/2 * sigma**2) + loc
        Var[X] = (exp(sigma**2)-1)*exp(2*mu + s**2)
        """
        RVG.__init__(self)
        self.mu = mu
        self.sigma = sigma
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.lognormal(mean=self.mu, sigma=self.sigma) + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.lognorm,
            params=[self.sigma, self.loc, np.exp(self.mu)])

    def pdf(self, x):
        return np.exp(-(np.log(x) - self.mu)**2 / (2 * self.sigma**2)) / (x * self.sigma * np.sqrt(2 * np.pi))

    def ppf(self, q):
        return stat.lognorm.ppf(q, s=self.sigma, loc=self.loc, scale=np.exp(self.mu))

    @staticmethod
    def fit_mm(mean, st_dev, fixed_location=0):
        """
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :param fixed_location: location, 0 by default
        :return: dictionary with keys "mu", "sigma" and "loc"
        """

        mean = mean-fixed_location

        mu = np.log(
            mean**2 / np.sqrt(st_dev**2 + mean**2)
        )
        sigma = np.sqrt(
            np.log(1 + st_dev**2 / mean**2)
        )

        return {"mu": mu, "sigma": sigma, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "mu", "sigma" and "loc"
        """

        # estimate the parameters
        s, loc, scale = stat.lognorm.fit(data, floc=fixed_location)
        sigma = s
        mu = np.log(scale)

        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.lognorm.logpdf(data, s, loc, scale)))

        # report results in the form of a dictionary
        return {"mu": mu, "sigma": sigma, "loc": loc, "AIC": aic}


class Multinomial(RVG):
    def __init__(self, N, pvals):
        """
        E[X_i] = N*p_i
        Var[X] = N*p_i(1-p_i)
        :param N: (int) number of trials
        :param pvals: (array) probabilities of success for each category
        """
        RVG.__init__(self)
        self.N = N
        self.pvals = pvals

    def sample(self, rng, arg=None):
        return rng.multinomial(self.N, self.pvals)


class NegativeBinomial(RVG):
    def __init__(self, n, p, loc=0):
        """
        The probability distribution for number of failure before n successes
        :param n: number of the number of successes
        :param p: p is the probability of a single success
        E[X] = (n*p)/(1-p) + loc
        Var[X] = (n*p)/((1-p)**2)
        """
        RVG.__init__(self)
        self.n = n
        self.p = p
        self.loc = loc

    def sample(self, rng, arg=None):
        """
        :return: a realization from the NegativeBinomial distribution
        (the number of failure before a specified number of successes, n, occurs.)
        """
        return rng.negative_binomial(self.n, self.p) + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.nbinom,
            params=[self.n, self.p, self.loc])

    @staticmethod
    def ln_likelihood(n, p, data):
        result = 0
        for d in data:
            result += stat.nbinom.logpmf(d, n, p)
        return result

    @staticmethod
    def fit_mm(mean, st_dev, fixed_location=0):
        """
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "n", "p" and "loc"
        """
        # in Scipy, n is the number of successes, p is the probability of a single success.
        # in Wiki, r is the number of failure, p is success probability
        # to match the moments, define r = n is the number of successes, 1-p is the failure probability
        mean = mean - fixed_location

        p = mean/st_dev**2.0
        n = mean*p/(1-p)

        return {"n": n, "p": p, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "n", "p" and "loc"
        """

        ## not working
        return

        data = data - fixed_location

        def neg_ln_l(n_p):
            return -NegativeBinomial.ln_likelihood(n=n_p[0], p=n_p[1], data=data)

        n_p_loc = NegativeBinomial.fit_mm(mean=np.mean(data), st_dev=np.std(data), fixed_location=0)
        n_p = [n_p_loc['n'], n_p_loc['p']]
        M = 2*np.max(data)

        # call Scipy optimizer to minimize the target function
        # with bounds for p [0,1] and n [0,M]
        fitted_n_p, value, iter, imode, smode = fmin_slsqp(neg_ln_l, n_p,
                                                           #bounds=[(0.0, M), (0, 1)],
                                                           disp=False, full_output=True)

        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.nbinom.logpmf(data, fitted_n_p[0], fitted_n_p[1], 0)))

        # report results in the form of a dictionary
        return {"n": fitted_n_p[0], "p": fitted_n_p[1], "loc": fixed_location, "AIC": aic}


class NonHomogeneousExponential(RVG):
    def __init__(self, rates, delta_t=None, time_breaks=None):
        """
        :param rates: (list) of rates over each period (e.g. [1, 2])
        :param delta_t: length of each period
        :param time_breaks: (list) time points marking the intervals
        Either delta_t or time_breaks should be provided, but not both.
        If delta_t is provided, then the time intervals are assumed to be
            [0, delta_t), [delta_t, 2*delta_t), ...
        If time_breaks is provided, then the time intervals are assumed to be
            [time_breaks[0], time_breaks[1]), [time_breaks[1], time_breaks[2]), ...
        Note that len(rates) = len(time_breaks) if time_breaks is provided,
            and len(rates) = len(time_breaks) - 1 if delta_t is provided.
        The rate of the last period should be greater than 0.
        """

        if delta_t is None and time_breaks is None:
            raise ValueError("Either delta_t or time_breaks should be provided.")
        if delta_t is not None and time_breaks is not None:
            raise ValueError("Only one of parameters delta_t or time_breaks should be provided.")

        if rates[-1] <= 0:
            raise ValueError('For a non-homogeneous exponential distribution, '
                             'the rate of the last period should be greater than 0.')

        RVG.__init__(self)
        self.rates = rates
        self.deltaT = delta_t

        if time_breaks is None:
            time_breaks = [i*delta_t for i in range(len(rates))]

        self.timeBreaks = time_breaks

    def sample(self, rng, arg=None):
        """
        :param arg: current time (age)
        :return: a realization from the NonHomogeneousExponential distribution
        """

        t = 0 # current time
        i = 0  # current interval

        if arg is not None:
            t = arg
            if self.deltaT is not None:
                i = min(math.floor(t/self.deltaT), len(self.rates)-1)
            else:
                found = False
                i = len(self.rates)-1
                while not found:
                    if t >= self.timeBreaks[i]:
                        found = True
                    else:
                        i -= 1

        if_occurred = False
        while not if_occurred:
            if self.rates[i] > 0:
                exp = Exponential(scale=1/self.rates[i])
                delta_t = exp.sample(rng)
            else:
                delta_t = float('inf')

            next_step = self.timeBreaks[i+1] if i < len(self.timeBreaks)-1 else float('inf')

            if t + delta_t < next_step:
                if_occurred = True
                t = t + delta_t
            else:
                t = self.timeBreaks[i+1]
                i += 1

        return t - arg


class Normal(RVG):
    def __init__(self, loc=0, scale=1):
        """
        E[X] = loc
        Var[X] = scale**2
        """
        RVG.__init__(self)
        self.loc = loc
        self.scale = scale

    def sample(self, rng, arg=None):
        return rng.normal(self.loc, self.scale)

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.norm,
            params=[self.loc, self.scale])

    @staticmethod
    def fit_mm(mean, st_dev):
        """
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :return: dictionary with keys "loc" and "scale"
        """

        return {"loc": mean, "scale": st_dev}

    @staticmethod
    def fit_ml(data):
        """
        :param data: (numpy.array) observations
        :return: dictionary with keys "loc", "scale", and "AIC"
        """

        mean = np.average(data)
        st_dev = np.std(data)
        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.norm.logpdf(data, mean, st_dev)))

        return {"loc": mean, "scale": st_dev, "AIC": aic}


class Poisson(RVG):
    def __init__(self, mu, loc=0):
        """
        E[X] = mu + loc
        Var[X] = mu
        """
        RVG.__init__(self)
        self.mu = mu
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.poisson(self.mu) + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.poisson,
            params=[self.mu, self.loc])

    @staticmethod
    def fit_mm(mean, fixed_location=0):
        """
        :param mean: sample mean of an observation set
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "mu" and "loc"
        """

        mu = mean - fixed_location

        return {"mu": mu, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "mu" and "loc"
        """

        # fit poisson distribution: the MLE of lambda is the sample mean
        # https://en.wikipedia.org/wiki/Poisson_distribution#Maximum_likelihood
        data = data - fixed_location
        mu = data.mean()

        # calculate AIC
        aic = AIC(
            k=1,
            log_likelihood=np.sum(stat.poisson.logpmf(data, mu)))

        # report results in the form of a dictionary
        return {"mu": mu, "loc": fixed_location, "AIC": aic}


class Triangular(RVG):
    def __init__(self, c, loc=0, scale=1):
        """
        l = loc, u = loc+scale, mode = loc + c*scale
        E[X] = (l+mode+u)/3
        Var[X] = (l**2 + mode**2 + u**2 -l*u - l*mode - u*mode)/18
        """
        RVG.__init__(self)
        self.c = c
        self.loc = loc
        self.scale = scale

    def sample(self, rng, arg=None):
        return stat.triang.rvs(self.c, self.loc, self.scale, random_state=rng)

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.triang,
            params=[self.c, self.loc, self.scale])

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :return: dictionary with keys "c", "loc", "scale", and "AIC"
        """

        c, loc, scale = stat.triang.fit(data, floc=fixed_location)

        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.triang.logpdf(data, c, loc, scale))
        )

        # report results in the form of a dictionary
        return {"c": c, "loc": loc, "scale": scale, "AIC": aic}


class Uniform(RVG):
    def __init__(self, scale=1, loc=0):
        """
        setting l = loc, u = loc + scale
        E[X] = (l+u)/2
        Var[X] = (u-l)**2/12
        """
        RVG.__init__(self)
        self.scale = scale
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.uniform(low=self.loc, high=self.loc+self.scale)

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.uniform,
            params=[self.loc, self.scale])

    @staticmethod
    def fit_mm(mean, st_dev):
        """
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :return: dictionary with keys "loc" and "scale"
        """

        b = 0.5*(2*mean + np.sqrt(12)*st_dev)
        a = 2.0*mean - b

        loc = a
        scale = b-a

        return {"loc": loc, "scale": scale,}

    @staticmethod
    def fit_ml(data):
        """
        :param data: (numpy.array) observations
        :return: dictionary with keys "loc", "scale", and "AIC"
        """

        # estimate the parameters
        loc, scale = stat.uniform.fit(data)

        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.uniform.logpdf(data, loc, scale))
        )

        # report results in the form of a dictionary
        return {"loc": loc, "scale": scale, "AIC": aic}


class UniformDiscrete(RVG):
    def __init__(self, l, u):
        """
        E[X] = (l+u)/2
        Var[X] = ((u-l+1)**2 - 1)/12
        :param l: (int) inclusive lower bound
        :param u: (int) inclusive upper bound
        """
        RVG.__init__(self)
        self.l = l
        self.u = u

    def sample(self, rng, arg=None):
        return rng.randint(low=self.l, high=self.u + 1)

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.randint,
            params=[self.l, self.u])

    @staticmethod
    def fit_mm(mean, st_dev):
        """
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :return: dictionary with keys "l" and "u"
        """
        variance = st_dev**2
        b = (np.sqrt(12.0*variance + 1) + 2.0*mean-1)*0.5
        a = (-np.sqrt(12.0*variance + 1) + 2.0*mean+1)*0.5

        return {"l": a, "u": b}

    @staticmethod
    def fit_ml(data):
        """
        :param data: (numpy.array) observations
        :return: dictionary with keys "l" and "u"
        """
        # estimate the parameters
        # as likelihood = 1/(high-low)^n, so the smaller the range, the higher the likelihood
        # the MLE is
        low = np.min(data)
        high = np.max(data)

        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.randint.logpmf(data, low, high))
        )

        # report results in the form of a dictionary
        return {"l": low, "u": high, "AIC": aic}


class Weibull(RVG):
    def __init__(self, a, scale=1, loc=0):
        """
        E[X] = math.gamma(1 + 1/a) * scale + loc
        Var[X] = [math.gamma(1 + 2/a) - (math.gamma(1 + 1/a)**2)] * scale**2
        """
        RVG.__init__(self)
        self.a = a
        self.scale = scale
        self.loc = loc

    def sample(self, rng, arg=None):
        return rng.weibull(self.a) * self.scale + self.loc

    def get_percentile_interval(self, alpha=0.05):

        return self._get_percentile_interval(
            alpha=alpha, dist=stat.weibull_min,
            params=[self.a, self.loc, self.scale])

    @staticmethod
    def fit_mm(mean, st_dev, fixed_location=0):
        """
        :param mean: sample mean of an observation set
        :param st_dev: sample standard deviation of an observation set
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "c", "loc" and "scale"
        """
        mean = mean - fixed_location

        c = (st_dev*1.0/mean)**(-1.086)
        scale = mean/sp.special.gamma(1 + 1.0/c)

        return {"c": c, "scale": scale, "loc": fixed_location}

    @staticmethod
    def fit_ml(data, fixed_location=0):
        """
        :param data: (numpy.array) observations
        :param fixed_location: location, 0 by default
        :returns: dictionary with keys "c", "loc" and "scale"
        """

        # estimate the parameters of weibull
        c, loc, scale = stat.weibull_min.fit(data, floc=fixed_location)

        # calculate AIC
        aic = AIC(
            k=2,
            log_likelihood=np.sum(stat.weibull_min.logpdf(data, c, loc, scale))
        )

        # report results in the form of a dictionary
        return {"c": c, "loc": loc, "scale": scale, "AIC": aic}


import math
import sys
import warnings

import numpy as np
import scipy.stats as stat
import statsmodels.api as sm
from scipy.stats import chi2
from scipy.stats import pearsonr
from statsmodels.stats.proportion import proportion_confint

import deampy.format_functions as F

NUM_BOOTSTRAP_SAMPLES = 1000


def remove_outliers(data, iq_factor=3.0):
    """
    :param data: (list) a list of numbers
    :param iq_factor: (float) number of inter quantiles from the median to consider as outlier
    :return: a list of numbers with outliers removed
    """

    if len(data) == 0:
        return []

    iqr = stat.iqr(data)
    q1 = np.percentile(data, 25)
    q3 = np.percentile(data, 75)
    u = q3 + iq_factor * iqr
    l = q1 - iq_factor * iqr

    return [d for d in data if l <= d <= u]


def get_sterr_from_half_length(confidence_interval, n, alpha=0.05):
    """
    :param confidence_interval: a t-based confidence interval
    :param n: number of observations
    :param alpha: significance level
    :return: the standard err (st_dev / sqrt(n))
    """
    half_length = (confidence_interval[1] - confidence_interval[0]) / 2
    t = stat.t.ppf(1 - alpha / 2, n - 1)
    return half_length / t


def partial_corr(x, y, z):
    """
    :param x: (list) of values for the input variable
    :param y: (list) of values for the output variable
    :param z: (list of list) list of values for the controlling variables
                [[z11, z12, ..., z1n],  # values of first controlling variable
                 [z21, z22, ..., z2n],  # values of second controlling variable
                 ...
                ]
    :return: partial correlation between x and y given the set of controlling variables z
    """

    np_z = np.array(z).T

    Z = sm.add_constant(np_z)
    # fit f(z) to x
    fitted_xz = sm.OLS(x, Z).fit()
    # fit f(z) to y
    fitted_yz = sm.OLS(y, Z).fit()
    # residuals f(z) - x
    x_residuals = fitted_xz.predict() - x
    # residuals f(z) - y
    y_residuals = fitted_yz.predict() - y
    # correlation between residuals
    coef, p = pearsonr(x_residuals, y_residuals)
    return coef, p


class _Statistics(object):
    def __init__(self, name=None):
        """ abstract method to be overridden in derived classes"""
        self.name = name  # name of this statistics
        self._n = 0  # number of data points
        self._mean = 0  # sample mean
        self._stDev = 0  # sample standard deviation
        self._max = -sys.float_info.max  # maximum
        self._min = sys.float_info.max  # minimum

    def get_mean(self):
        """ abstract method to be overridden in derived classes
        :returns mean (to be calculated in the subclass) """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_stdev(self):
        """ abstract method to be overridden in derived classes
        :returns standard deviation (to be calculated in the subclass) """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_stdev_CI(self, alpha):
        """ abstract method to be overridden in derived classes
        :returns the confidence interval for standard deviation (to be calculated in the subclass) """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_var(self):
        return self.get_stdev() ** 2

    def get_min(self):
        """ abstract method to be overridden in derived classes
        :returns minimum (to be calculated in the subclass) """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_max(self):
        """ abstract method to be overridden in derived classes
        :returns maximum (to be calculated in the subclass) """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_percentile(self, q):
        """ abstract method to be overridden in derived classes
        :param q: percentile to compute (q in range [0, 100])
        :returns percentile (to be calculated in the subclass) """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_t_half_length(self, alpha):
        """
        :param alpha: significance level (between 0 and 1)
        :returns half-length of 100(1-alpha)% t-confidence interval """

        if self._n > 1:
            return stat.t.ppf(1 - alpha / 2, self._n - 1) * self.get_stdev() / np.sqrt(self._n)
        else:
            return math.nan

    def get_t_CI(self, alpha):
        """ calculates t-based confidence interval for population mean
        :param alpha: significance level (between 0 and 1)
        :return: a list [l, u]
        """

        if self._n > 1:
            mean = self.get_mean()
            hl = self.get_t_half_length(alpha)
            return [mean - hl, mean + hl]
        else:
            return [math.nan, math.nan]

    def _get_proportion_CI(self, data, alpha):
        """
        :param alpha: significance level (between 0 and 1)
        :return: Wilon score interval in the format of list [l, u]
        """

        # observations should be either 0 or 1
        for d in data:
            if d not in (0, 1):
                raise ValueError('Wilson score interval can only be calculated for binary (0 or 1) outcomes.')
        return proportion_confint(count=sum(data), nobs=self._n, alpha=alpha, method='wilson')

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """ calculates empirical bootstrap confidence interval (abstract method to be overridden in derived classes)
        :param alpha: significance level
        :param num_samples: number of bootstrap samples
        :returns a list [L, U] """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_PI(self, alpha):
        """ calculates percentile interval (abstract method to be overridden in derived classes)
        :param alpha: significance level (between 0 and 1)
        :returns a list [L, U]
         """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")

    def get_summary(self, alpha, digits):
        """
        :param alpha: significance level
        :param digits: digits to round the numbers to
        :return: a list ['name', 'mean', 'confidence interval', 'percentile interval', 'st dev', 'min', 'max']
        """
        return [self.name,
                F.format_number(self.get_mean(), digits),
                F.format_interval(self.get_t_CI(alpha), digits),
                F.format_interval(self.get_PI(alpha), digits),
                F.format_number(self.get_stdev(), digits),
                F.format_number(self.get_min(), digits),
                F.format_number(self.get_max(), digits)]

    def get_interval(self, interval_type='c', mean_or_stdev='mean',
                     alpha=0.05, multiplier=1, num_of_bootstrap_samples=None):
        """
        :param interval_type: (string) 'c' for t-based confidence interval,
                                       'cb' for bootstrap confidence interval,
                                       'p' for percentile interval, and
                                       'w' for Wilson score interval (only for binary outcomes)
        :param mean_or_stdev: 'mean' or 'stdev' to indicate whether the estimate is mean or standard deviation
        :param alpha: significance level
        :param multiplier: to multiply the estimate and the interval by the provided value
        :param num_of_bootstrap_samples: number of bootstrap samples (only needed if interval_type is 'cb')
        :return: a list [L, U]
        """

        if mean_or_stdev == 'mean':
            if interval_type == 'c':
                interval = self.get_t_CI(alpha)
            elif interval_type == 'cb':
                if num_of_bootstrap_samples is None:
                    num_of_bootstrap_samples = NUM_BOOTSTRAP_SAMPLES
                interval = self.get_bootstrap_CI(alpha, num_of_bootstrap_samples)
            elif interval_type == 'p':
                interval = self.get_PI(alpha)
            elif interval_type == 'w':
                if isinstance(self, (SummaryStat, DifferenceStatPaired)):
                    interval = self.get_proportion_CI(alpha)
                else:
                    raise ValueError(
                        'Wilson score interval can only be calculated for SummaryStat or DifferenceStatPaired.')
            elif interval_type == 'n' or interval_type is None:
                interval = None
            else:
                raise ValueError('Invalid interval type.')
            if interval is not None:
                return [v * multiplier for v in interval]
            else:
                return None

        elif mean_or_stdev == 'stdev':
            if interval_type in ('c', 'w'):
                interval = self.get_stdev_CI(alpha)
                if interval is not None:
                    return [v * multiplier for v in interval]
                else:
                    return None
            elif interval_type == 'n' or interval_type is None:
                return None
            else:
                raise ValueError('Invalid interval type.')
        else:
            raise ValueError("mean_or_stdev should be either 'mean' or 'stdev'.")

    def get_formatted_mean_and_interval(self, interval_type='c', mean_or_stdev='mean',
                                        alpha=0.05, deci=None, sig_digits=None, form=None, multiplier=1):
        """
        :param interval_type: (string) 'c' for t-based confidence interval,
                                       'cb' for bootstrap confidence interval, and
                                       'p' for percentile interval
        :param mean_or_stdev: 'mean' or 'stdev' to indicate whether the estimate is mean or standard deviation
        :param alpha: significance level
        :param deci: digits to round the numbers to
        :param sig_digits: number of significant digits
        :param form: ',' to format as number, '%' to format as percentage, and '$' to format as currency
        :param multiplier: to multiply the estimate and the interval by the provided value
        :return: (string) estimate and interval formatted as specified
        """

        if mean_or_stdev == 'mean':
            estimate = self.get_mean() * multiplier
        elif mean_or_stdev == 'stdev':
            estimate = self.get_stdev() * multiplier
        else:
            raise ValueError("mean_or_stdev should be either 'mean' or 'stdev'.")

        interval = self.get_interval(
            interval_type=interval_type, mean_or_stdev=mean_or_stdev,
            alpha=alpha, multiplier=multiplier)

        return F.format_estimate_interval(
            estimate=estimate, interval=interval,
            deci=deci, sig_digits=sig_digits, format=form)

    def get_formatted_interval(self, interval_type='c',
                               alpha=0.05, deci=None, sig_digits=None, form=None, multiplier=1):
        """
        :param interval_type: (string) 'c' for t-based confidence interval,
                                       'cb' for bootstrap confidence interval, and
                                       'p' for percentile interval
        :param alpha: significance level
        :param deci: digits to round the numbers to
        :param sig_digits: number of significant digits
        :param form: ',' to format as number, '%' to format as percentage, and '$' to format as currency
        :param multiplier: to multiply the estimate and the interval by the provided value
        :return: (string) interval formatted as specified
        """

        interval = self.get_interval(interval_type=interval_type, alpha=alpha, multiplier=multiplier)

        return F.format_interval(interval=interval,
                                 deci=deci,
                                 sig_digits=sig_digits,
                                 format=form)


class SummaryStat(_Statistics):
    def __init__(self, data, name=None):
        """:param data: a list or numpy.array of data points"""

        _Statistics.__init__(self, name)
        # convert data to numpy array if needed
        if type(data) is list:
            self._data = np.array(data)
        elif type(data) is np.ndarray:
            self._data = data
        else:
            raise ValueError("The argument data can be either a list of numbers or a numpy.array.")

        if len(data) == 0:
            raise ValueError("The data provided for summary statistics '" + name + "' is empty.")

        self._n = len(self._data)
        self._total = np.sum(self._data)
        self._mean = np.mean(self._data)
        if self._n > 1:
            self._stDev = np.std(self._data, ddof=1)  # unbiased estimator of the standard deviation
        else:
            self._stDev = math.nan

    def get_total(self):
        return self._total

    def get_mean(self):
        return self._mean

    def get_stdev(self):
        return self._stDev

    def get_stdev_CI(self, alpha=0.05):

        n = len(self._data)
        df = n - 1
        s2 = np.var(self._data, ddof=1)  # unbiased sample variance

        # Chi-square critical values
        chi2_lower = chi2.ppf(alpha / 2, df)
        chi2_upper = chi2.ppf(1 - alpha / 2, df)

        lower = math.sqrt((df * s2) / chi2_upper)
        upper = math.sqrt((df * s2) / chi2_lower)

        return [lower, upper]

    def get_min(self):
        return np.min(self._data)

    def get_max(self):
        return np.max(self._data)

    def get_percentile(self, q):
        """
        :param q: percentile to compute (q in range [0, 100])
        :returns: qth percentile """

        return float(np.percentile(self._data, q))

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """ calculates the empirical bootstrap confidence interval
        :param alpha: significance level (between 0 and 1)
        :param num_samples: number of bootstrap samples
        :return: a list [l, u]
        """

        if num_samples is None:
            num_samples = NUM_BOOTSTRAP_SAMPLES

        # set random number generator seed
        rng = np.random.RandomState(1)

        # initialize delta array
        delta = np.zeros(num_samples)

        # obtain bootstrap samples
        for i in range(num_samples):
            sample_i = rng.choice(self._data, size=self._n, replace=True)
            delta[i] = sample_i.mean() - self.get_mean()

        # return [l, u]
        return self.get_mean() - np.percentile(delta, [100 * (1 - alpha / 2.0), 100 * alpha / 2.0])

    def get_proportion_CI(self, alpha=0.05):
        """
        :param alpha: significance level (between 0 and 1)
        :return: Wilson score interval in the format of list [l, u]
        """

        return self._get_proportion_CI(data=self._data, alpha=alpha)

    def get_PI(self, alpha=0.05):
        """
        :param alpha: significance level (between 0 and 1)
        :return: percentile interval in the format of list [l, u]
        """
        return [self.get_percentile(100 * alpha / 2), self.get_percentile(100 * (1 - alpha / 2))]

    @staticmethod
    def get_array_from_formatted_interval(interval):
        """
        :param interval: (string) of form '[a, b]'
        :return: (np.array) [a, b]
        """
        str_range = interval[1:-1].split(',')
        return np.array(str_range).astype(float)

    @staticmethod
    def get_mean_interval_from_formatted_mean_interval(formatted_text):
        """
        :param formatted_text: (string) of form 'a (b, c)'
        :param form: ','  '%' or '$'
        :return: (float, np.array) a, [b, c]
        """

        # remove characters
        for char in ('$', '%', ',', '(', ')'):
            formatted_text = formatted_text.replace(char, '')

        # split by space and convert to float
        array = np.array(formatted_text.split(' ')).astype(float)

        return array[0], array[1:]


class DiscreteTimeStat(_Statistics):
    """ to calculate statistics on observations accumulating over time """

    def __init__(self, name=None):
        _Statistics.__init__(self, name)
        self._total = 0
        self._sumSquared = 0

    def record(self, obs):
        """ gets the next observation and update the current information"""
        self._total += obs
        self._sumSquared += obs ** 2
        self._n += 1
        if obs > self._max:
            self._max = obs
        if obs < self._min:
            self._min = obs

    def get_total(self):
        return self._total

    def get_mean(self):
        if self._n > 0:
            return self._total / self._n
        else:
            return 0

    def get_stdev(self):
        if self._n > 1:
            return math.sqrt(
                (self._sumSquared - self._total ** 2 / self._n)
                / (self._n - 1)
            )
        else:
            return 0

    def get_min(self):
        return self._min

    def get_max(self):
        return self._max

    def get_percentile(self, q):
        """ percentiles cannot be calculated for this statistics """
        return None

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """ bootstrap confidence intervals cannot be calculated for this statistics """
        return None

    def get_PI(self, alpha):
        """ percentile intervals cannot be calculated for this statistics """
        return None


class ContinuousTimeStat(_Statistics):
    """ to calculate statistics on the area-under-the-curve for observations accumulating over time """

    def __init__(self, initial_time, ave_method='step', name=None):
        """
        :param initial_time: it is assumed that the value of this sample path is zero at the initial time
        :param ave_method: to calculate the area under the curve,
            'step' assumes that changes occurred at the time where observations are recorded,
            'linear' assumes uses trapezoid approach to calculate the area under the curve
                    (v2 + v2 ) * deltaT / 2
        """
        _Statistics.__init__(self, name)
        self._initialTime = initial_time

        if ave_method not in ('step', 'linear'):
            raise ValueError('ave_method should be either step or linear.')
        self._aveMethod = ave_method
        self._area = 0
        self._areaSquared = 0
        self._lastObsTime = initial_time
        self._lastObsValue = 0

    def record(self, time, increment):
        """ gets the next observation and update the current information
        :param time: time of this change in the sample path
        :param increment: the amount of change (can be positive or negative) in the sample path
        """

        if time < self._initialTime:
            self._lastObsTime = time
            self._lastObsValue += increment
            return
        else:
            self._lastObsTime = max(self._lastObsTime, self._initialTime)

        if self._lastObsValue > self._max:
            self._max = self._lastObsValue
        if time == self._initialTime:
            self._min = self._lastObsValue
        elif self._lastObsValue < self._min:
            self._min = self._lastObsValue

        self._n += 1
        if self._aveMethod == 'step':
            self._area += self._lastObsValue * (time - self._lastObsTime)
            self._areaSquared += (self._lastObsValue ** 2) * (time - self._lastObsTime)
        else:
            self._area += (self._lastObsValue + increment / 2) * (time - self._lastObsTime)
            self._areaSquared += ((self._lastObsValue + increment / 2) ** 2) * (time - self._lastObsTime)

        self._lastObsTime = time
        self._lastObsValue += increment

    def get_mean(self):
        if self._lastObsTime - self._initialTime > 0:
            return self._area / (self._lastObsTime - self._initialTime)
        else:
            return 0

    def get_stdev(self):
        var = 0
        if self._lastObsTime - self._initialTime > 0:
            var = self._areaSquared / (self._lastObsTime - self._initialTime) - self.get_mean() ** 2

        return math.sqrt(var)

    def get_min(self):
        return self._min

    def get_max(self):
        return self._max

    def get_percentile(self, q):
        """ percentiles cannot be calculated for this statistics """
        return None

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """ bootstrap confidence intervals cannot be calculated for this statistics """
        return None

    def get_PI(self, alpha):
        """ percentile intervals cannot be calculated for this statistics """
        return None


class _ComparativeStat(_Statistics):
    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations (the reference) 
        """
        _Statistics.__init__(self, name)

        if isinstance(x, list):
            self._x = np.array(x)
        elif isinstance(x, np.ndarray):
            self._x = x
        else:
            raise ValueError("The argument x can be either a list of numbers or a numpy.array.")

        if isinstance(y_ref, list):
            self._y_ref = np.array(y_ref)
        elif isinstance(y_ref, np.ndarray):
            self._y_ref = y_ref
        else:
            raise ValueError("The argument y_ref can be either a list of numbers or a numpy.array.")

        if len(self._x) == 0:
            raise ValueError("x is empty for the comparative statistics '" + name + "'.")
        if len(self._y_ref) == 0:
            raise ValueError("y_ref is empty for the comparative statistics '" + name + "'.")

        self._x_n = len(self._x)  # number of observations for x
        self._y_n = len(self._y_ref)  # number of observations for y_ref


class _DifferenceStat(_ComparativeStat):

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        """
        _ComparativeStat.__init__(self, x, y_ref, name)


class DifferenceStatPaired(_DifferenceStat):

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        """
        _DifferenceStat.__init__(self, x, y_ref, name)
        # create a summary statistics for the element-wise difference

        if len(self._x) != len(self._y_ref):
            raise ValueError('Two samples (x and y_ref) should have the same size.')

        self._dStat = SummaryStat(self._x - self._y_ref, name)

    def get_mean(self):
        return self._dStat.get_mean()

    def get_stdev(self):
        return self._dStat.get_stdev()

    def get_min(self):
        return self._dStat.get_min()

    def get_max(self):
        return self._dStat.get_max()

    def get_percentile(self, q):
        return self._dStat.get_percentile(q)

    def get_t_CI(self, alpha):
        return self._dStat.get_t_CI(alpha)

    def get_bootstrap_CI(self, alpha, num_samples=None):
        return self._dStat.get_bootstrap_CI(alpha, num_samples)

    def get_proportion_CI(self, alpha=0.05):
        return self._get_proportion_CI(data=self._x - self._y_ref, alpha=alpha)

    def get_PI(self, alpha):
        return self._dStat.get_PI(alpha)


class DifferenceStatIndp(_DifferenceStat):

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        """
        _DifferenceStat.__init__(self, x, y_ref, name)

        self._XMinusYSimulated = False

    def _simulate_x_minus_y(self):

        # generate random realizations for random variable X - Y
        # this will be used for calculating the projection interval
        rng = np.random.RandomState(1)
        # find the maximum of the number of observations
        max_n = max(self._x_n, self._y_n, NUM_BOOTSTRAP_SAMPLES)
        x_i = rng.choice(self._x, size=max_n, replace=True)
        y_i = rng.choice(self._y_ref, size=max_n, replace=True)
        self._sum_stat_sample_delta = SummaryStat(x_i - y_i, self.name)

        self._XMinusYSimulated = True

    def get_mean(self):
        """
        for independent variable x and y, E(x-y) = E(x) - E(y)
        :return: sample mean of (x-y)
        """
        return np.mean(self._x) - np.mean(self._y_ref)

    def get_stdev(self):
        """
        for independent variable x and y, var(x-y) = var_x + var_y
        :returns: sample standard deviation
        """
        var_x = np.var(self._x)
        var_y = np.var(self._y_ref)
        return np.sqrt(var_x + var_y)

    def get_min(self):

        if not self._XMinusYSimulated:
            self._simulate_x_minus_y()
        return self._sum_stat_sample_delta.get_min()

    def get_max(self):

        if not self._XMinusYSimulated:
            self._simulate_x_minus_y()
        return self._sum_stat_sample_delta.get_max()

    def get_percentile(self, q):
        """
        for independent variable x and y, percentiles are given after re-sampling
        :param q: the percentile want to return, in [0, 100]
        :return: qth percentile of sample (x-y)
        """

        if not self._XMinusYSimulated:
            self._simulate_x_minus_y()

        return self._sum_stat_sample_delta.get_percentile(q)

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """
        :param alpha: confidence level
        :param num_samples: number of bootstrap samples
        :return: empirical bootstrap confidence interval
        """

        n = NUM_BOOTSTRAP_SAMPLES if num_samples is None else num_samples

        # set random number generator seed
        rng = np.random.RandomState(1)

        # find number of bootstrap samples
        n = max(self._x_n, self._y_n, n)

        # initialize difference array
        diff = np.zeros(n)

        # obtain bootstrap samples
        for i in range(n):
            x_i = rng.choice(self._x, size=n, replace=True)
            y_i = rng.choice(self._y_ref, size=n, replace=True)
            d_temp = x_i - y_i
            diff[i] = np.mean(d_temp)

        return np.percentile(diff, [100 * alpha / 2.0, 100 * (1 - alpha / 2.0)])

    def get_t_half_length(self, alpha):
        """
        Independent x_bar - y_bar is t distribution
        ref: https://onlinecourses.science.psu.edu/stat414/node/203
        :param alpha: confidence level
        :return: confidence interval of x_bar - y_bar
        """

        sig_x = np.std(self._x)
        sig_y = np.std(self._y_ref)

        alpha = alpha / 100.0

        # calculate CI using formula: Welch's t-interval
        df_n = (sig_x ** 2.0 / self._x_n + sig_y ** 2.0 / self._y_n) ** 2.0
        df_d = (sig_x ** 2.0 / self._x_n) ** 2 / (self._x_n - 1) \
               + (sig_y ** 2.0 / self._y_n) ** 2 / (self._y_n - 1)
        df = round(df_n / df_d, 0)

        # t distribution quantile
        t_q = stat.t.ppf(1 - (alpha / 2), df)
        st_dev = (sig_x ** 2.0 / self._x_n + sig_y ** 2.0 / self._y_n) ** 0.5

        return t_q * st_dev

    def get_t_CI(self, alpha):

        interval = self.get_t_half_length(alpha)
        diff = np.mean(self._x) - np.mean(self._y_ref)

        return [diff - interval, diff + interval]

    def get_PI(self, alpha):

        if not self._XMinusYSimulated:
            self._simulate_x_minus_y()

        return self._sum_stat_sample_delta.get_PI(alpha)


class _RatioStat(_ComparativeStat):

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        """
        _ComparativeStat.__init__(self, x, y_ref, name)

        self._ifComputable = True  # if any element of the denominator is 0, ratio is not computable


class RatioStatPaired(_RatioStat):
    # E[X/Y] when X and Y are paired

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        """
        _RatioStat.__init__(self, x, y_ref, name)

        if len(self._x) != len(self._y_ref):
            raise ValueError('Two samples (x and y_ref) should have the same size.')

        # add element-wise ratio
        self.ratios = np.zeros(len(self._x))
        for i in range(len(self._x)):
            # for 0 in the denominator variable, check whether numerator is also 0
            if self._y_ref[i] == 0:
                self.ratios[i] = math.nan
                warnings.warn("For ratio statistics '{}', "
                              "the denominator of ratio with index {} is 0."
                              "The ratio is not computable.".format(name, i))
                self._ifComputable = False
            else:
                # for non-zero denominators, calculate ratio
                self.ratios[i] = 1.0 * self._x[i] / self._y_ref[i]

        # create summary stat for element-wise ratio
        self._ratioStat = SummaryStat(self.ratios, name)

    def get_mean(self):
        if self._ifComputable:
            return self._ratioStat.get_mean()
        else:
            return math.nan

    def get_stdev(self):
        if self._ifComputable:
            return self._ratioStat.get_stdev()
        else:
            return math.nan

    def get_min(self):
        return self._ratioStat.get_min()

    def get_max(self):
        return self._ratioStat.get_max()

    def get_percentile(self, q):
        if self._ifComputable:
            return self._ratioStat.get_percentile(q)
        else:
            return math.nan

    def get_t_CI(self, alpha):
        if self._ifComputable:
            return self._ratioStat.get_t_CI(alpha)
        else:
            return math.nan

    def get_bootstrap_CI(self, alpha, num_samples=None):
        if self._ifComputable:
            return self._ratioStat.get_bootstrap_CI(alpha, num_samples)
        else:
            return math.nan

    def get_PI(self, alpha):
        if self._ifComputable:
            return self._ratioStat.get_PI(alpha)
        else:
            return [math.nan, math.nan]


class RatioOfMeansStatPaired(_RatioStat):
    # E[X]/E[Y] when X and Y are paired

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        """
        _RatioStat.__init__(self, x, y_ref, name)

        if len(self._x) != len(self._y_ref):
            raise ValueError('Two samples (x and y_ref) should have the same size.')

        # add element-wise ratio
        self.ratios = np.zeros(len(self._x))
        for i in range(len(self._x)):
            # for 0 in the denominator variable, check whether numerator is also 0
            if self._y_ref[i] == 0:
                self.ratios[i] = math.nan
                warnings.warn("For ratio statistics '{}', "
                              "the denominator of ratio with index {} is 0."
                              "The confidence or prediction interval will not be computable.".format(name, i))
                self._ifComputable = False
            else:
                # for non-zero denominators, calculate ratio
                self.ratios[i] = 1.0 * self._x[i] / self._y_ref[i]

    def get_mean(self):
        return np.mean(self._x) / np.mean(self._y_ref)

    def get_stdev(self):
        return np.nan

    def get_min(self):
        return np.nan

    def get_max(self):
        return np.nan

    def get_percentile(self, q):
        return np.nan

    def get_t_CI(self, alpha):
        return [np.nan, np.nan]

    def get_bootstrap_CI(self, alpha, num_samples=None):

        if num_samples is None:
            num_samples = NUM_BOOTSTRAP_SAMPLES

        # set random number generator seed
        rng = np.random.RandomState(1)

        # initialize delta array
        delta = np.zeros(num_samples)

        mean = self.get_mean()

        # obtain bootstrap samples
        for i in range(num_samples):
            i_s = rng.choice(range(len(self._x)), size=len(self._x), replace=True)
            x_i = self._x[i_s]
            y_i = self._y_ref[i_s]
            delta[i] = np.mean(x_i) / np.mean(y_i) - mean

        # return [l, u]
        return mean - np.percentile(delta, [100 * (1 - alpha / 2.0), 100 * alpha / 2.0])

    def get_PI(self, alpha):
        return [np.nan, np.nan]


class RatioStatIndp(_RatioStat):
    # E[X/Y] when X and Y are independent

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations (reference)

        https://en.wikipedia.org/wiki/Ratio_distribution
        """

        _RatioStat.__init__(self, x, y_ref, name)

        self._XOverYSimulated = False

    def _simulate_x_over_y(self):

        # generate random realizations for random variable X/Y
        rng = np.random.RandomState(1)

        # find the maximum of the number of observations
        max_n = max(self._x_n, self._y_n, NUM_BOOTSTRAP_SAMPLES)
        x_resample = rng.choice(self._x, size=max_n, replace=True)
        y_resample = rng.choice(self._y_ref, size=max_n, replace=True)

        self._XOverYSimulated = True
        self._ifComputable = True
        try:
            self._sum_stat_sample_ratio = SummaryStat(
                data=np.divide(x_resample, y_resample), name=self.name)
        except ZeroDivisionError:
            warnings.warn("For ratio statistics '{}', "
                          "one element of y_ref is 0.".format(self.name))
            self._ifComputable = False

    def get_mean(self):
        """
        for independent variable x and y, E(x/y) = E(x)*E(1/y)
        :return: E(x/y)
        """

        return np.average(self._x) * np.average(1 / self._y_ref)

    def get_stdev(self):
        """
        for independent variable x and y, var(x/y) = E(x^2)*E(1/y^2)-E(x)^2*(E(1/y)^2)
        :return: std(x/y)
        """
        var = np.mean(self._x ** 2) * np.mean(1.0 / self._y_ref ** 2) - \
              (np.mean(self._x) ** 2) * (np.mean(1.0 / self._y_ref) ** 2)
        return np.sqrt(var)

    def get_min(self):

        if not self._XOverYSimulated:
            self._simulate_x_over_y()
        if self._ifComputable:
            return self._sum_stat_sample_ratio.get_min()
        else:
            return math.nan

    def get_max(self):

        if not self._XOverYSimulated:
            self._simulate_x_over_y()
        if self._ifComputable:
            return self._sum_stat_sample_ratio.get_max()
        else:
            return math.nan

    def get_percentile(self, q):
        """
        for independent variable x and y, percentiles are given after re-sampling
        :param q: the percentile want to return, in [0,100]
        :return: qth percentile of sample (x/y)
        """

        if not self._XOverYSimulated:
            self._simulate_x_over_y()
        if self._ifComputable:
            return self._sum_stat_sample_ratio.get_percentile(q)
        else:
            return math.nan

    def get_t_half_length(self, alpha):
        """ cannot be calculated for ratio statistics """

        return math.nan

        # if not self._XOverYSimulated:
        #     self._simulate_x_over_y()
        # if self._ifComputable:
        #     return self._sum_stat_sample_ratio.get_t_half_length(alpha)
        # else:
        #     return math.nan

    def get_t_CI(self, alpha):
        """ cannot be calculated for ratio statistics """

        return [math.nan, math.nan]

        # if self._x_n > 1 and self._y_n > 1:
        #     mean = self.get_mean()
        #     hl = self.get_t_half_length(alpha)
        #     return [mean - hl, mean + hl]
        # else:
        #     return [math.nan, math.nan]

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """
        :param alpha: confidence level
        :param num_samples: number of samples
        :return: empirical bootstrap confidence interval
        """

        if num_samples is None:
            num_samples = NUM_BOOTSTRAP_SAMPLES

        # set random number generator seed
        rng = np.random.RandomState(1)

        # initialize delta array
        delta = np.zeros(num_samples)

        # obtain bootstrap samples
        n = max(self._x_n, self._y_n)
        mean = self.get_mean()
        for i in range(num_samples):
            x_i = rng.choice(self._x, size=n, replace=True)
            y_i = rng.choice(self._y_ref, size=n, replace=True)
            ratios_i = np.average(x_i) * np.average(1 / y_i)
            delta[i] = ratios_i - mean

        # return [l, u]
        return mean - np.percentile(delta, [100 * (1 - alpha / 2.0), 100 * alpha / 2.0])

    def get_PI(self, alpha):

        if not self._XOverYSimulated:
            self._simulate_x_over_y()

        if self._ifComputable:
            return self._sum_stat_sample_ratio.get_PI(alpha)
        else:
            return [math.nan, math.nan]


class RatioOfMeansStatIndp(_RatioStat):
    # E[X]/E[Y] when X and Y are independent

    def __init__(self, x, y_ref, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        """
        _RatioStat.__init__(self, x, y_ref, name)

    def get_mean(self):
        """
        :return: the ratio of the means of x and y_ref"""
        return np.mean(self._x) / np.mean(self._y_ref)

    def get_stdev(self):
        return np.nan

    def get_min(self):
        return np.nan

    def get_max(self):
        return np.nan

    def get_percentile(self, q):
        return np.nan

    def get_t_CI(self, alpha):
        return [np.nan, np.nan]

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """
         :param alpha: confidence level
         :param num_samples: number of samples
         :return: empirical bootstrap confidence interval
         """

        if num_samples is None:
            num_samples = NUM_BOOTSTRAP_SAMPLES

        # set random number generator seed
        rng = np.random.RandomState(1)

        # initialize delta array
        delta = np.zeros(num_samples)

        # obtain bootstrap samples
        n = max(self._x_n, self._y_n)
        mean = self.get_mean()
        for i in range(num_samples):
            x_i = rng.choice(self._x, size=n, replace=True)
            y_i = rng.choice(self._y_ref, size=n, replace=True)
            ratios_i = np.average(x_i) / np.average(y_i)
            delta[i] = ratios_i - mean

        # return [l, u]
        return mean - np.percentile(delta, [100 * (1 - alpha / 2.0), 100 * alpha / 2.0])

    def get_PI(self, alpha):
        return [np.nan, np.nan]


class _RelativeDifference(_ComparativeStat):
    """ class to make inference about (X-Y_ref)/Y_ref or (Y_ref-X)/Y_ref"""

    def __init__(self, x, y_ref, order=0, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations used as the reference values
        :param order: set to 0 to calculate (X-Y_ref)/Y and to 1 to calculate (Y_ref-X)/Y_ref
        """
        _ComparativeStat.__init__(self, x, y_ref, name)
        self._order = order
        self._ifComputable = True  # if any element of the denominator is 0, ratio is not computable


class RelativeDifferencePaired(_RelativeDifference):
    # E[(X-Y)/Y] when X and Y are paired

    def __init__(self, x, y_ref, order=0, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        :param order: set to 0 to calculate (X-Y_ref)/Y_ref and to 1 to calculate (Y_ref-X)/Y_ref
        """
        _RelativeDifference.__init__(self, x, y_ref, order, name)

        if len(self._x) != len(self._y_ref):
            raise ValueError('Two samples should have the same size.')

        # add element-wise ratio
        ratio = np.zeros(len(self._x))

        for i in range(len(self._x)):
            # for 0 in the denominator variable, check whether numerator is also 0
            if self._y_ref[i] == 0:
                ratio[i] = math.nan
                warnings.warn("For ratio statistics '{}', "
                              "the denominator of ratio with index {} is 0."
                              "The ratio is not computable.".format(name, i))
                self._ifComputable = False
            # for non-zero denominators, calculate ratio
            else:
                ratio[i] = 1.0 * self._x[i] / self._y_ref[i]

        # create summary stat for element-wise ratio
        if self._order == 0:
            self._relativeDiffStat = SummaryStat(ratio - 1, name)
        else:
            self._relativeDiffStat = SummaryStat(1 - ratio, name)

    def get_mean(self):
        if self._ifComputable:
            return self._relativeDiffStat.get_mean()
        else:
            return math.nan

    def get_stdev(self):
        if self._ifComputable:
            return self._relativeDiffStat.get_stdev()
        else:
            return math.nan

    def get_min(self):
        return self._relativeDiffStat.get_min()

    def get_max(self):
        return self._relativeDiffStat.get_max()

    def get_percentile(self, q):
        if self._ifComputable:
            return self._relativeDiffStat.get_percentile(q)
        else:
            return math.nan

    def get_t_CI(self, alpha):
        if self._ifComputable:
            return self._relativeDiffStat.get_t_CI(alpha)
        else:
            return [math.nan, math.nan]

    def get_bootstrap_CI(self, alpha, num_samples=None):

        if num_samples is None:
            num_samples = NUM_BOOTSTRAP_SAMPLES

        if self._ifComputable:
            return self._relativeDiffStat.get_bootstrap_CI(alpha, num_samples)
        else:
            return [math.nan, math.nan]

    def get_PI(self, alpha):
        if self._ifComputable:
            return self._relativeDiffStat.get_PI(alpha)
        else:
            return [math.nan, math.nan]


class RelativeDiffOfMeansPaired(_RelativeDifference):
    # (E[X]-E[Y])/E[Y] when X and Y are paired

    def __init__(self, x, y_ref, order=0, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        :param order: set to 0 to calculate (X-Y_ref)/Y_ref and to 1 to calculate (Y_ref-X)/Y_ref
        """
        _RelativeDifference.__init__(self, x, y_ref, order, name)

        if len(self._x) != len(self._y_ref):
            raise ValueError('Two samples should have the same size.')

        self._ratioOfMeans = RatioOfMeansStatPaired(x, y_ref, name)

    def get_mean(self):
        if self._order == 0:
            return self._ratioOfMeans.get_mean() - 1
        else:
            return 1 - self._ratioOfMeans.get_mean()

    def get_stdev(self):
        return np.nan

    def get_min(self):
        return np.nan

    def get_max(self):
        return np.nan

    def get_percentile(self, q):
        return np.nan

    def get_t_CI(self, alpha):
        return [np.nan, np.nan]

    def get_bootstrap_CI(self, alpha, num_samples=None):

        interval = self._ratioOfMeans.get_bootstrap_CI(alpha, num_samples)
        if self._order == 0:
            return [interval[0] - 1, interval[1] - 1]
        else:
            return [1 - interval[1], 1 - interval[0]]

    def get_PI(self, alpha):
        return [np.nan, np.nan]


class RelativeDifferenceIndp(_RelativeDifference):
    # E[(X-Y)/Y] when X and Y are independent

    def __init__(self, x, y_ref, order=0, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        :param order: set to 0 to calculate (X-Y_ref)/Y_ref and to 1 to calculate (Y_ref-X)/Y_ref
        """
        _RelativeDifference.__init__(self, x, y_ref, order, name)

        self._ratioStat = RatioStatIndp(x, y_ref, name)

    def get_mean(self):
        """
        for independent variable x and y, E((x-y)/y) = E(x)*E(1/y) - 1
        :return: E((x-y)/y) or E((y-x)/y)
        """
        if self._order == 0:
            return self._ratioStat.get_mean() - 1
        else:
            return 1 - self._ratioStat.get_mean()

    def get_stdev(self):
        """
        for independent variable x and y, var(x/y) = E(x^2)*E(1/y^2)-E(x)^2*(E(1/y)^2)
        and var(x/y - 1) = var(x/y)
        :return: std(x/y - 1)
        """
        return self._ratioStat.get_stdev()

    def get_min(self):

        if self._order == 0:
            return self._ratioStat.get_min() - 1
        else:
            return 1 - self._ratioStat.get_max()

    def get_max(self):

        if self._order == 0:
            return self._ratioStat.get_max() - 1
        else:
            return 1 - self._ratioStat.get_min()

    def get_percentile(self, q):
        """
        for independent variable x and y, percentiles are given after re-sampling
        :param q: the percentile want to return, in [0,100]
        :return: qth percentile of sample (x-y)/y
        """

        if self._order == 0:
            return self._ratioStat.get_percentile(q) - 1
        else:
            return 1 - self._ratioStat.get_percentile(100 - q)

    def get_t_half_length(self, alpha):
        """ cannot be calculated for ratio statistics """

        return math.nan

    def get_t_CI(self, alpha):
        """ cannot be calculated for ratio statistics """

        return [math.nan, math.nan]

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """
        :param alpha: confidence level
        :param num_samples: number of samples
        :return: empirical bootstrap confidence interval
        """

        if num_samples is None:
            num_samples = NUM_BOOTSTRAP_SAMPLES

        interval = self._ratioStat.get_bootstrap_CI(alpha, num_samples)

        if self._order == 0:
            return [interval[0] - 1, interval[0] - 1]
        else:
            return [1 - interval[1], 1 - interval[0]]

    def get_PI(self, alpha):

        interval = self._ratioStat.get_PI(alpha)
        if self._order == 0:
            return [interval[0] - 1, interval[1] - 1]
        else:
            return [1 - interval[1], 1 - interval[0]]


class RelativeDiffOfMeansIndp(_RelativeDifference):
    # (E[X]-E[Y])/E[Y] when X and Y are independent

    def __init__(self, x, y_ref, order=0, name=None):
        """
        :param x: list or numpy.array of first set of observations
        :param y_ref: list or numpy.array of second set of observations
        :param order: set to 0 to calculate (X-Y_ref)/Y_ref and to 1 to calculate (Y_ref-X)/Y_ref
        """
        _RelativeDifference.__init__(self, x, y_ref, order, name)

        self._ratioOfMeans = RatioOfMeansStatIndp(x, y_ref, name)

    def get_mean(self):
        """
        :return: the relative difference of means of x and y_ref"""
        if self._order == 0:
            return self._ratioOfMeans.get_mean() - 1
        else:
            return 1 - self._ratioOfMeans.get_mean()

    def get_stdev(self):
        return np.nan

    def get_min(self):
        return np.nan

    def get_max(self):
        return np.nan

    def get_percentile(self, q):
        return np.nan

    def get_t_CI(self, alpha):
        return [np.nan, np.nan]

    def get_bootstrap_CI(self, alpha, num_samples=None):
        """
         :param alpha: confidence level
         :param num_samples: number of samples
         :return: empirical bootstrap confidence interval
         """

        interval = self._ratioOfMeans.get_bootstrap_CI(alpha, num_samples)
        if self._order == 0:
            return [interval[0] - 1, interval[1] - 1]
        else:
            return [1 - interval[1], 1 - interval[0]]

    def get_PI(self, alpha):
        return [np.nan, np.nan]
